# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..base import torrent as base
from . import _mixin as mixin
from .. import ExpectedResource
from ._common import make_tab_title_widget

from functools import partial


class ListTorrentsCmd(base.ListTorrentsCmdbase,
                      mixin.select_torrents,
                      mixin.create_list_widget):
    provides = {'tui'}

    def make_torrent_list(self, tfilter, sort, columns):
        from ...tui.views.torrent_list import TorrentListWidget
        self.create_list_widget(TorrentListWidget, theme_name='torrentlist',
                                tfilter=tfilter, sort=sort, columns=columns,
                                markable_items=True)


class TorrentSummaryCmd(base.TorrentSummaryCmdbase,
                        mixin.select_torrents, mixin.make_request):
    provides = {'tui'}
    tui = ExpectedResource

    async def display_summary(self, torrent_id):
        make_titlew = partial(make_tab_title_widget,
                              attr_unfocused='tabs.torrentsummary.unfocused',
                              attr_focused='tabs.torrentsummary.focused')

        from ...tui.views.summary import TorrentSummaryWidget
        TorrentSummaryWidget_keymapped = self.tui.keymap.wrap(TorrentSummaryWidget,
                                                              context='torrent')
        title_str = self.title if hasattr(self, 'title') else None
        summaryw = TorrentSummaryWidget_keymapped(self.srvapi, torrent_id, title=title_str)
        tabid = self.tui.tabs.load(make_titlew(summaryw.title), summaryw)

        def set_tab_title(text):
            # set_title() throws IndexError if the tab was removed, which may
            # have happened while TorrentSummaryWidget was waiting for a
            # response.
            try:
                self.tui.tabs.set_title(make_titlew(text), position=tabid)
            except IndexError:
                pass
        summaryw.title_updater = set_tab_title


class AddTorrentsCmd(base.AddTorrentsCmdbase,
                     mixin.polling_frenzy, mixin.make_request):
    provides = {'tui'}


class MoveTorrentsCmd(base.MoveTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class RenameTorrentCmd(base.RenameTorrentCmdbase,
                       mixin.polling_frenzy, mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'tui'}


class RemoveTorrentsCmd(base.RemoveTorrentsCmdbase,
                        mixin.polling_frenzy, mixin.make_request, mixin.select_torrents,
                        mixin.ask_yes_no):
    provides = {'tui'}
    cmdmgr = ExpectedResource
    CONFIRMATION_TAB_TITLE = 'Removal Confirmation'

    async def show_list_of_hits(self, tfilter):
        cmd = 'tab --title %r ls --sort name %s' % (self.CONFIRMATION_TAB_TITLE, tfilter)
        await self.cmdmgr.run_async(cmd)

    async def remove_list_of_hits(self):
        cmd = 'tab --close %r --focus left' % self.CONFIRMATION_TAB_TITLE
        await self.cmdmgr.run_async(cmd)

class StartTorrentsCmd(base.StartTorrentsCmdbase,
                       mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StopTorrentsCmd(base.StopTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class VerifyTorrentsCmd(base.VerifyTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'tui'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Torrent class and value modifiers for compatibility with ttypes"""

from ...logging import make_logger
log = make_logger(__name__)

from .. import ttypes
from ..utils import LazyDict
from .. import base


# Some values need to be modified to comply with our internal standards

def _modify_ratio(t):
    #define TR_RATIO_NA  -1
    #define TR_RATIO_INF -2
    ratio = t['uploadRatio']
    if ratio == -1:
        return ttypes.Ratio.NOT_APPLICABLE
    elif ratio == -2:
        return ttypes.Ratio.INFINITE
    else:
        return ratio


def _modify_eta(t):
    #define TR_ETA_NOT_AVAIL -1
    #define TR_ETA_UNKNOWN -2
    seconds = t['eta']
    if seconds == -1:
        return ttypes.Timedelta.NOT_APPLICABLE
    elif seconds == -2:
        return ttypes.Timedelta.UNKNOWN
    else:
        return seconds


def _modify_timestamp(t, key, zero_means=ttypes.Timestamp.UNKNOWN):
    # I couldn't find any documentation on this, but 0 seems to mean "not applicable"?
    seconds = t[key]
    if seconds == 0:
        return zero_means
    else:
        return seconds


import time
def _modify_timestamp_completed(t):
    if t['percentDone'] >= 1:
        doneDate = t['doneDate']
        if doneDate == 0:
            return t['addedDate']            # We are the original uploader
        else:
            return doneDate                  # Torrent has been completed in the past
    elif t['eta'] <= 0:
        return ttypes.Timestamp.UNKNOWN      # Torrent is incomplete + paused
    else:
        return time.time() + t['eta']        # Torrent is downloading


def _count_seeds(t):
    trackerStats = t['trackerStats']
    if trackerStats:
        return max(t['seederCount'] for t in trackerStats)
    else:
        return ttypes.Count.UNKNOWN


def _bytes_available(t):
    return t['desiredAvailable'] + t['haveValid'] + t['haveUnchecked']


def _percent_available(t):
    try:
        return _bytes_available(t) / t['sizeWhenDone'] * 100
    except ZeroDivisionError:
        return 0


def _percent_uploaded(t):
    try:
        return t['uploadedEver'] / t['totalSize'] * 100
    except ZeroDivisionError:
        return 0


def _is_isolated(t):
    """Return whether this torrent can find any peers via trackers or DHT"""
    if not t['isPrivate']:
        return False  # DHT is used

    # Torrent has trackers?
    trackerStats = t['trackerStats']
    if trackerStats:
        # Did we try to connect to a tracker?
        if any(tracker['hasAnnounced'] for tracker in trackerStats):
            # Did at least one tracker respond?
            if any(tracker['lastAnnounceSucceeded'] for tracker in trackerStats):
                return False
        # We didn't try yet; assume non-isolation
        else:
            return False
    return True  # No way to find any peers


def _find_error(t):
    error = t['error']
    if error == 1:
        return 'Tracker warning: %s' % t['errorString']
    elif error == 2:
        return 'Tracker error: %s' % t['errorString']
    elif error == 3:
        return t['errorString']

    # The fields 'error' and 'errorString' are not necessarily set when
    # _is_isolated returns True. (Not sure why that happens. Reproduce by
    # setting a tracker domain to 127.0.0.1 in /etc/hosts to provoke an error.)
    trackerStats = t['trackerStats']
    for tracker in trackerStats:
        msg = tracker['lastAnnounceResult']
        if msg != 'Success':
            return msg

    return ''


def _make_status(t):
    Status = ttypes.Status
    statuses = []

    # RPC values for 'status' field:
    # TR_STATUS_STOPPED        = 0, /* Torrent is stopped */
    # TR_STATUS_CHECK_WAIT     = 1, /* Queued to check files */
    # TR_STATUS_CHECK          = 2, /* Checking files */
    # TR_STATUS_DOWNLOAD_WAIT  = 3, /* Queued to download */
    # TR_STATUS_DOWNLOAD       = 4, /* Downloading */
    # TR_STATUS_SEED_WAIT      = 5, /* Queued to seed */
    # TR_STATUS_SEED           = 6  /* Seeding */
    t_status = t['status']
    if t_status == 0:
        statuses.append(Status.STOPPED)
    elif t_status in (1, 2):
        statuses.append(Status.VERIFY)
    if t_status in (1, 3, 5):
        statuses.append(Status.QUEUED)

    if Status.STOPPED not in statuses:
        if _is_isolated(t):
            statuses.append(Status.ISOLATED)
        if t['metadataPercentComplete'] < 1:
            statuses.append(Status.INIT)

        if Status.QUEUED not in statuses:
            if t['peersConnected'] > 0:
                if t['rateDownload'] > 0:
                    statuses.append(Status.DOWNLOAD)
                if t['rateUpload'] > 0:
                    statuses.append(Status.UPLOAD)
                statuses.append(Status.CONNECTED)

            if t['percentDone'] >= 1:
                statuses.append(Status.SEED)

    if all(x not in statuses for x in (Status.UPLOAD,
                                       Status.DOWNLOAD,
                                       Status.VERIFY)):
        statuses.append(Status.IDLE)

    return statuses


class TorrentFileID(tuple):
    def __new__(cls, torrent_id, file_id):
        return super().__new__(cls, (torrent_id, file_id))

    @property
    def torrent_id(self):
        return self[0]

    @property
    def file_id(self):
        return self[1]

    def __repr__(self):
        return 'TorrentFileID(torrent_id=%d, file_id=%d)' % self

import os
class TorrentFileTree(base.TorrentFileTreeBase):
    @classmethod
    def create(cls, raw_torrent):
        fileStats = raw_torrent['fileStats']
        if len(fileStats) < 1:
            # filelist is empty if torrent was added by hash and metadata isn't
            # downloaded yet.
            filelist = [{'tid': -1, 'id': TorrentFileID(-1, -1), 'name': raw_torrent['name'],
                         'priority': 0, 'length': 0, 'wanted': True, 'bytesCompleted': 0}]
        else:
            # Combine 'files' and 'fileStats' fields and add the 'id' key to each
            # file, which is a (torrent ID, file list index) tuple
            tid = raw_torrent['id']
            filelist = ({'id': TorrentFileID(tid, i), **f, **fS}
                        for i,(f,fS) in enumerate(zip(raw_torrent['files'], fileStats)))
        return cls(raw_torrent['id'], raw_torrent['downloadDir'], filelist, path=())

    def __init__(self, torrent_id, torrent_location, filelist, path):
        log.debug('Creating new TorrentFileTree for torrent %r: %r', torrent_id, path)
        path_str = os.sep.join(path)
        super().__init__(torrent_location, path_str)

        items = {}
        subdirs = {}
        for entry in filelist:
            parts = entry['name'].split(os.sep, 1)
            if len(parts) == 1:
                filename = parts[0]
                items[filename] = ttypes.TorrentFile(
                    tid=torrent_id, id=entry['id'],
                    name=entry['name'], path=path_str, location=torrent_location,
                    size_total=entry['length'],
                    size_downloaded=entry['bytesCompleted'],
                    is_wanted=entry['wanted'],
                    priority=entry['priority'])

            elif len(parts) == 2:
                subdir, subpath = parts
                if subdir not in subdirs:
                    subdirs[subdir] = []
                entry['name'] = subpath
                subdirs[subdir].append(entry)
            else:
                raise RuntimeError(parts)

        for subdir,filelist in subdirs.items():
            items[subdir] = TorrentFileTree(torrent_id, torrent_location,
                                            filelist, path=path+(subdir,))
        self._items = items

    def update(self, raw_torrent):
        def update_files(ftree, fileStats):
            if not fileStats:
                # We don't have any metadata yet, so there is nothing to update
                return

            for entry in ftree.values():
                if isinstance(entry, ttypes.TorrentFile):
                    # File ID is its index in the list provided by
                    # Transmission (see _create_TorrentFileTree)
                    index = entry['id'].file_id
                    fstats = fileStats[index]
                    entry.update({'size-downloaded': fstats['bytesCompleted'],
                                  'is-wanted': fstats['wanted'],
                                  'priority': fstats['priority'],
                                  'location': raw_torrent['downloadDir']})
                else:
                    update_files(entry, fileStats)

        update_files(self._items, raw_torrent['fileStats'])



class PeerList(tuple):
    def __new__(cls, t):
        TorrentPeer = ttypes.TorrentPeer
        return super().__new__(cls,
            (TorrentPeer(tid=t['id'], tname=t['name'], tsize=t['totalSize'],
                         ip=p['address'], port=p['port'], client=p['clientName'],
                         pdownloaded=p['progress']*100,
                         rate_up=p['rateToPeer'], rate_down=p['rateToClient'])
             for p in t['peers'])
        )



class TrackerList(tuple):
    _STATES_ANNOUNCE = {
        # From libtransmission/transmission.h:
        # /* we won't (announce,scrape) this torrent to this tracker because
        #  * the torrent is stopped, or because of an error, or whatever */
        0: 'stopped',
        # /* we will (announce,scrape) this torrent to this tracker, and are
        #  * waiting for enough time to pass to satisfy the tracker's interval */
        1: 'idle',
        # /* it's time to (announce,scrape) this torrent, and we're waiting on a
        #  * a free slot to open up in the announce manager */
        2: 'queued',
        # /* we're (announcing,scraping) this torrent right now */
        3: 'announcing',
    }
    _STATES_SCRAPE = {0: 'stopped', 1: 'idle', 2: 'queued', 3: 'scraping'}

    @staticmethod
    def _error_announce(tracker):
        msg = tracker['lastAnnounceResult'] if tracker['hasAnnounced'] else ''
        return '' if msg == 'Success' else msg

    @staticmethod
    def _error_scrape(tracker):
        msg = tracker['lastScrapeResult'] if tracker['hasScraped'] else ''
        return '' if msg == 'Success' else msg

    @staticmethod
    def _next_time(tracker, which):
        """
        Handle next(Announce|Scrape)Time RPC key

        `which` must be 'Scrape' or 'Announce'.

        transmission.h says:
            /* when the next periodic (announce|scrape) message will be sent out.
               if (announce|scrape)State isn't TR_TRACKER_WAITING, this field is undefined */
        """
        state = tracker['%sState' % which.lower()]
        if state == 1:    # TR_TRACKER_WAITING = 1
            return tracker['next%sTime' % which]
        elif state == 0:  # Torrent is paused
            return ttypes.Timestamp.NOT_APPLICABLE
        elif state == 2:  # Announce/scrape is queued
            return ttypes.Timestamp.SOON
        else:
            return ttypes.Timestamp.NOW

    @staticmethod
    def _last_time(tracker, which):
        """
        Handle last(Announce|Scrape)Time RPC key

        `which` must be 'Scrape' or 'Announce'.

        transmission.h says:
            /* when the last (announce|scrape) was completed.
               if "has(Announced|Scraped)" is false, this field is undefined */
        """
        if tracker['has%sd' % which]:
            return tracker['last%sTime' % which]
        else:
            return ttypes.Timestamp.NEVER

    def __new__(cls, raw_torrent):
        return super().__new__(cls,
            (ttypes.TorrentTracker(
                (LazyDict({
                    'id'                 : (raw_torrent['id'], raw_tracker['id']),
                    'tid'                : raw_torrent['id'],
                    'tname'              : raw_torrent['name'],
                    'tier'               : raw_tracker['tier'],

                    'url-announce'       : raw_tracker['announce'],
                    'url-scrape'         : raw_tracker['scrape'],

                    'status-announce'    : cls._STATES_ANNOUNCE[raw_tracker['announceState']],
                    'status-scrape'      : cls._STATES_SCRAPE[raw_tracker['scrapeState']],

                    'error-announce'     : lambda: cls._error_announce(raw_tracker),
                    'error-scrape'       : lambda: cls._error_scrape(raw_tracker),

                    'count-downloads'    : raw_tracker['downloadCount'],
                    'count-leeches'      : raw_tracker['leecherCount'],
                    'count-seeds'        : raw_tracker['seederCount'],

                    'time-last-announce' : lambda: cls._last_time(raw_tracker, 'Announce'),
                    'time-last-scrape'   : lambda: cls._last_time(raw_tracker, 'Scrape'),
                    'time-next-announce' : lambda: cls._next_time(raw_tracker, 'Announce'),
                    'time-next-scrape'   : lambda: cls._next_time(raw_tracker, 'Scrape'),
                }))) for raw_tracker in raw_torrent['trackerStats'])
        )



# Map our keys to tuples of needed RPC field names for those keys
DEPENDENCIES = {
    'id'                           : ('id',),
    'hash'                         : ('hashString',),
    'name'                         : ('name',),
    'ratio'                        : ('uploadRatio',),
    'status'                       : ('status', 'percentDone', 'metadataPercentComplete', 'rateDownload',
                                      'rateUpload', 'peersConnected', 'trackerStats', 'isPrivate'),
    'path'                         : ('downloadDir',),
    'private'                      : ('isPrivate',),
    'comment'                      : ('comment',),
    'creator'                      : ('creator',),
    'magnetlink'                   : ('magnetLink',),
    'count-pieces'                 : ('pieceCount',),

    '%downloaded'                  : ('percentDone',),
    '%uploaded'                    : ('totalSize', 'uploadedEver'),
    '%metadata'                    : ('metadataPercentComplete',),
    '%verified'                    : ('recheckProgress',),
    '%available'                   : ('haveValid', 'haveUnchecked', 'desiredAvailable', 'sizeWhenDone'),

    'peers-connected'              : ('peersConnected',),
    'peers-uploading'              : ('peersSendingToUs',),
    'peers-downloading'            : ('peersGettingFromUs',),
    'peers-seeding'                : ('trackerStats',),

    'timespan-eta'                 : ('eta',),
    'timespan-seeding'             : ('secondsSeeding',),
    'timespan-downloading'         : ('secondsDownloading',),
    'time-created'                 : ('dateCreated',),
    'time-added'                   : ('addedDate',),
    'time-started'                 : ('startDate',),
    'time-activity'                : ('activityDate',),
    'time-completed'               : ('doneDate', 'addedDate', 'percentDone', 'eta'),
    'time-manual-announce-allowed' : ('manualAnnounceTime',),

    'rate-down'                    : ('rateDownload',),
    'rate-up'                      : ('rateUpload',),
    'limit-rate-down'              : ('downloadLimited', 'downloadLimit'),
    'limit-rate-up'                : ('uploadLimited', 'uploadLimit'),

    'size-final'                   : ('sizeWhenDone',),
    'size-total'                   : ('totalSize',),
    'size-downloaded'              : ('downloadedEver',),
    'size-uploaded'                : ('uploadedEver',),
    'size-available'               : ('leftUntilDone', 'desiredAvailable', 'haveValid', 'haveUnchecked'),
    'size-left'                    : ('leftUntilDone',),
    'size-corrupt'                 : ('corruptEver',),
    'size-piece'                   : ('pieceSize',),

    'error'                        : ('errorString', 'error', 'trackerStats'),
    'trackers'                     : ('trackerStats', 'name', 'id'),
    'peers'                        : ('peers', 'totalSize', 'name'),
    'files'                        : ('files', 'fileStats', 'downloadDir'),
}

# Map our keys to callables that adjust the raw RPC values or create new
# values from existing RPC values.
_MODIFY = {
    '%downloaded'                  : lambda raw: raw['percentDone'] * 100,
    '%uploaded'                    : _percent_uploaded,
    '%metadata'                    : lambda raw: raw['metadataPercentComplete'] * 100,
    '%verified'                    : lambda raw: raw['recheckProgress'] * 100,
    '%available'                   : _percent_available,
    'status'                       : _make_status,
    'peers-seeding'                : _count_seeds,
    'ratio'                        : _modify_ratio,
    'size-available'               : _bytes_available,

    # Transmission provides rate limits in kilobytes - we want bytes
    'limit-rate-down'              : lambda raw: None if not raw['downloadLimited'] else raw['downloadLimit'] * 1000,
    'limit-rate-up'                : lambda raw: None if not raw['uploadLimited']   else raw['uploadLimit']   * 1000,

    'timespan-eta'                 : _modify_eta,
    'time-created'                 : lambda raw: _modify_timestamp(raw, 'dateCreated',
                                                                   zero_means=ttypes.Timestamp.UNKNOWN),
    'time-added'                   : lambda raw: _modify_timestamp(raw, 'addedDate',
                                                                   zero_means=ttypes.Timestamp.UNKNOWN),
    'time-started'                 : lambda raw: _modify_timestamp(raw, 'startDate',
                                                                   zero_means=ttypes.Timestamp.NOT_APPLICABLE),
    'time-activity'                : lambda raw: _modify_timestamp(raw, 'activityDate',
                                                                   zero_means=ttypes.Timestamp.NEVER),
    'time-completed'               : lambda raw: _modify_timestamp_completed(raw),
    'time-manual-announce-allowed' : lambda raw: _modify_timestamp(raw, 'manualAnnounceTime',
                                                                   zero_means=ttypes.Timestamp.NEVER),

    'error'                        : _find_error,
    'trackers'                     : TrackerList,
    'peers'                        : PeerList,
    'files'                        : TorrentFileTree.create,
}

class Torrent(base.TorrentBase):
    """
    Information about a torrent as a mapping

    The available keys are specified in DEPENDENCIES and ttypes.TYPES.
    """

    def __init__(self, raw_torrent):
        self._raw = raw_torrent
        self._cache = {}

    def update(self, raw_torrent):
        cache = self._cache
        raw_old = self._raw

        # Remove cached values if their original/raw value(s) differ
        for k in tuple(cache):
            # Each key depends on one or more RPC field
            fields = DEPENDENCIES[k]
            for field in fields:
                new_value = raw_torrent.get(field)
                old_value = raw_old.get(field)
                if new_value is not None and new_value != old_value:
                    # log.debug('Invalidating cached %s/%s: %r -> %r', k, field, old_value, new_value)
                    # New and previous value differ - if we are dealing with
                    # more complex data structures (e.g. a file tree), use the
                    # update() method to update the object in cache instead of
                    # removing it from the cache.
                    value = cache[k]
                    if hasattr(value, 'update') and all(field in raw_torrent for field in fields):
                        value.update(raw_torrent)
                    del cache[k]
                    break

        # Now we can forget the old values
        raw_old.update(raw_torrent)

    def __getitem__(self, key):
        cache = self._cache
        if key not in cache:
            raw = self._raw
            if key in _MODIFY:
                # Modifier gets the whole raw torrent
                value = _MODIFY[key](raw)
            else:
                # Copy raw value unmodified
                fields = DEPENDENCIES[key]
                value = raw[fields[0]]
            cache[key] = ttypes.TYPES[key](value)
        return cache[key]

    def __contains__(self, key):
        deps = DEPENDENCIES
        raw = self._raw
        # Check if key is known
        if key not in deps:
            return False
        else:
            # Check if we have all dependencies for key
            for dep in deps[key]:
                if dep not in raw:
                    return False
        return True

    def __iter__(self):
        for key in DEPENDENCIES:
            if key in self:
                yield key

    def __eq__(self, other):
        if hasattr(other, '_raw'):
            return self._raw['id'] == other._raw['id']
        else:
            return NotImplemented

    def __lt__(self, other):
        if hasattr(other, '_raw'):
            return self._raw['id'] > other._raw['id']
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self._raw['id'])

    def clearcache(self):
        self._cache = {}


class TorrentFields(tuple):
    """
    Convert Torrent keys to those specified in rpc-spec.txt

    The resulting tuple has no duplicates and the keys 'id' and 'name' are
    always included.
    """
    _RPC_FIELDS = ('activityDate', 'addedDate', 'announceResponse', 'announceURL',
                   'bandwidthPriority', 'comment', 'corruptEver', 'creator',
                   'dateCreated', 'desiredAvailable', 'doneDate', 'downloadDir',
                   'downloadedEver', 'downloadLimit', 'downloadLimited',
                   'downloadLimitMode', 'error', 'errorString', 'eta', 'etaIdle',
                   'hashString', 'haveUnchecked', 'haveValid', 'honorsSessionLimits',
                   'id', 'isFinished', 'isPrivate', 'isStalled', 'lastAnnounceTime',
                   'lastScrapeTime', 'leftUntilDone', 'magnetLink',
                   'manualAnnounceTime', 'maxConnectedPeers',
                   'metadataPercentComplete', 'name', 'nextAnnounceTime',
                   'nextScrapeTime', 'peer-limit', 'peersConnected',
                   'peersGettingFromUs', 'peersSendingToUs', 'percentDone',
                   'pieceCount', 'pieceSize', 'queuePosition', 'rateDownload',
                   'rateUpload', 'recheckProgress', 'secondsDownloading',
                   'secondsSeeding', 'scrapeResponse', 'scrapeURL', 'seedIdleLimit',
                   'seedIdleMode', 'seedRatioLimit', 'seedRatioMode', 'sizeWhenDone',
                   'startDate', 'status', 'totalSize', 'torrentFile', 'uploadedEver',
                   'uploadLimit', 'uploadLimitMode', 'uploadLimited', 'uploadRatio',
                   'webseedsSendingToUs',

                   # Lists ('files' is handled internally in api_torrent -
                   # request 'fileStats' instead)
                   'fileStats', 'peers', 'peersFrom', 'pieces', 'priorities',
                   'trackers', 'trackerStats', 'wanted', 'webseeds')

    _ALL_FIELDS = tuple(set(field
                            for fields in DEPENDENCIES.values()
                            for field in fields))
    _cache = {}

    def __new__(cls, *keys):
        if keys not in cls._cache:
            cls._cache[keys] = super().__new__(cls, cls._get_fields(*keys))
        return cls._cache[keys]

    @classmethod
    def _get_fields(cls, *keys):
        collected_fields = set(('id',))
        for key in keys:
            if key.lower() == 'all':
                return cls._ALL_FIELDS
            elif key in DEPENDENCIES:
                # key is one of Torrent's keys that needs one or more RPC fields
                collected_fields.update(DEPENDENCIES[key])
            elif key in cls._RPC_FIELDS:
                # key is a valid Transmission RPC field
                collected_fields.add(key)
            else:
                raise ValueError('Unknown torrent key: {!r}'.format(key))
        return collected_fields

    def __add__(self, other):
        if isinstance(other, (type(self), set, list, tuple)):
            fields = set(self)  # Make a copy
            fields.update(other)
            return type(self)(*fields)
        else:
            return NotImplemented

    def __eq__(self, other):
        return set(self) == set(other)

    def __ne__(self, other):
        return not self.__eq__(other)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..base import config as base
from . import _mixin as mixin


class RcCmd(base.RcCmdbase):
    provides = {'tui'}

class ResetCmd(base.ResetCmdbase):
    provides = {'tui'}

    def run(self, NAME):
        if NAME:
            super().run(NAME)
        else:
            # Get name from focused item in setting list
            from ...tui import main as tui
            from ...tui.views.setting_list import SettingListWidget
            widget = tui.tabs.focus
            if isinstance(widget, SettingListWidget):
                setting_name = widget.focused_widget.name
                super().run((setting_name,))


class SetCmd(base.SetCmdbase,
             mixin.create_list_widget):
    provides = {'tui'}

    def make_setting_list(self, sort, columns):
        from ...tui.views.setting_list import SettingListWidget
        self.create_list_widget(SettingListWidget, theme_name='settinglist',
                                sort=sort, columns=columns)


class RateLimitCmd(base.RateLimitCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.polling_frenzy):
    provides = {'tui'}

    async def _set_limits(self, TORRENT_FILTER, directions, limit, adjust=False, quiet=False):
        if TORRENT_FILTER == ['global']:
            await self._set_global_limits(directions, limit,
                                          adjust=adjust, quiet=quiet)
        else:
            await self._set_individual_limits(TORRENT_FILTER, directions, limit,
                                              adjust=adjust, quiet=quiet)

    async def _show_limits(self, TORRENT_FILTER, directions):
        if TORRENT_FILTER == ['global']:
            await self._show_global_limits(directions)
        else:
            await self._show_individual_limits(TORRENT_FILTER, directions)

    def _output(self, msg):
        self.info(msg)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Commands that work exclusively in the TUI"""

from ...logging import make_logger
log = make_logger(__name__)

from .. import (InitCommand, CmdError, ExpectedResource)
from . import _mixin as mixin
from ._common import make_tab_title_widget

import shlex
from functools import partial
import os


# Import tui.main module only on demand
def _get_keymap_contexts():
    from ...tui.main import keymap
    return tuple(keymap.contexts)


class BindCmd(metaclass=InitCommand):
    name = 'bind'
    provides = {'tui'}
    category = 'tui'
    description = 'Bind keys to commands or other keys'
    usage = ('bind [<OPTIONS>] <KEY> <ACTION>',)
    examples = ('bind ctrl-a tab ls active',
                'bind --context tabs alt-[ tab --focus left',
                'bind --context tabs alt-] tab --focus right',
                'bind --context torrent alt-! start --force',
                "bind --context torrent 'd .' rm",
                "bind --context torrent 'd+!' rm --delete-files",
                'bind u <up>',
                'bind d <down>')
    argspecs = (
        { 'names': ('--context','-c'),
          'description': 'Where KEY is grabbed (see CONTEXTS section)' },
        { 'names': ('--description','-d'),
          'description': 'Explanation of what ACTION does' },
        { 'names': ('KEY',),
          'description': 'One or more keys or key combinations (see KEYS section)' },
        { 'names': ('ACTION',), 'nargs': 'REMAINDER',
          'description': ("Any command or '<KEY>' (including the brackets) "
                          'to translate one key to another') },
    )

    def __create_CONTEXTS_section():
        lines = [
            ('Keys are mapped in contexts.  With no context given, the default '
             'context is used.  The default context only gets the key if no '
             'other context wants it.  The same key can be mapped to different '
             'actions in different contexts.'),
            '',
            'Available contexts are: ' +
              ', '.join('%s' % context for context in _get_keymap_contexts()),
            '',
            'EXAMPLE',
            '\tbind --context torrent ctrl-t start',
            '\tbind --context tabs ctrl-t tab',
            '\tbind ctrl-t <left>',
            '',
            ('\tWhen focusing a torrent, <ctrl-t> starts the focused torrent.  '
             'If focus is not on a torrent but still on a tab (e.g. when reading '
             'documentation) a new tab is opened.  Otherwise (e.g. focus is on the '
             'command prompt), <ctrl-t> does the same as <left>.'),
        ]
        return lines

    more_sections = {
        'CONTEXTS': __create_CONTEXTS_section,
        'KEYS': (
            'Single-character keys are specified as themselves (e.g. h, X, 5, !, þ, ¥, etc).',
            '',
            ('Special key names are enter, space, tab, backspace, insert, delete, home, end, '
             'up, down, left, right, pgup, pgdn and f1-12.'),
            '',
            ("The modifiers 'ctrl', 'alt' and 'shift' are separated with '-' from the key "
             "(e.g. alt-i, shift-delete, ctrl-a).  shift-x is identical to X."),
            '',
            ("Chained keys are sparated by single spaces (' ') or pluses ('+') and must be "
             "given as a single argument."),
        )
    }

    tui = ExpectedResource

    def run(self, context, description, KEY, ACTION):
        keymap = self.tui.keymap

        key = KEY
        if len(ACTION) == 1 and ACTION[0][0] == '<' and ACTION[0][-1] == '>':
            # ACTION is another key (e.g. 'j' -> 'down')
            action = keymap.mkkey(ACTION[0])
        else:
            action = ' '.join(shlex.quote(x) for x in ACTION)

        if context is None:
            context = keymap.DEFAULT_CONTEXT
        elif context not in _get_keymap_contexts():
            raise CmdError('Invalid context: %r' % (context,))

        try:
            keymap.bind(key, action, context=context, description=description)
        except ValueError as e:
            raise CmdError(e)


class UnbindCmd(metaclass=InitCommand):
    name = 'unbind'
    provides = {'tui'}
    category = 'tui'
    description = 'Unbind keys so pressing them has no effect'
    usage = ('unbind [<OPTIONS>] <KEY> <KEY> ...',)
    examples = ('unbind --context main ctrl-l',
                'unbind q')
    argspecs = (
        { 'names': ('--context','-c'),
          'description': 'Where KEY is grabbed (see "bind" command)' },
        { 'names': ('--all','-a'), 'action': 'store_true',
          'description': 'Remove all existing keybindings, including defaults' },
        { 'names': ('KEY',), 'nargs': 'REMAINDER',
          'description': 'Keys or key combinations (see "bind" command)' },
    )

    tui = ExpectedResource

    def run(self, context, all, KEY):
        keymap = self.tui.keymap

        if all:
            keymap.clear()

        if context is None:
            context = keymap.DEFAULT_CONTEXT
        elif context not in _get_keymap_contexts():
            raise CmdError('Invalid context: %r' % (context,))

        success = True
        for key in KEY:
            try:
                keymap.unbind(key, context=context)
            except ValueError as e:
                self.error(e)
                success = False
            else:
                success = success and True
        if not success:
            raise CmdError()


class SetCommandCmd(mixin.placeholders, metaclass=InitCommand):
    name = 'setcommand'
    aliases = ('setcmd',)
    provides = {'tui'}
    category = 'tui'
    description = 'Open the command line and insert a command'
    usage = ('setcommand [--trailing-space] <COMMAND> <ARGUMENT> <ARGUMENT> ...',)
    examples = (
        'setcommand --trailing-space tab ls',
        '\tAsk the user for a filter before opening a new torrent list.',
        '',
        'setcommand move {{location}}/',
        '\tMove the focused torrent, using the current path as default.',
        '',
        'setcommand move id={{id}} {{location}}/',
        ('\tSame as above, but make sure to move the correct torrent in case '
         'it is removed from the list while typing in the new path, e.g. if '
         'we\'re listing active torrents and the focused torrent stops being active.'),
    )
    argspecs = (
        { 'names': ('COMMAND',), 'nargs': '+',
          'description': 'Command the can user edit before executing it (see PLACEHOLDERS)' },
        { 'names': ('--trailing-space',), 'action': 'store_true',
          'description': 'Append a space at the end of COMMAND' },
    )
    more_sections = {
        'PLACEHOLDERS': mixin.placeholders.HELP,
    }

    tui = ExpectedResource

    async def run(self, COMMAND, trailing_space):
        log.debug('Unresolved command: %r', COMMAND)
        args = await self.parse_placeholders(*COMMAND)
        log.debug('Command with resolved placeholders: %r', args)

        if args:
            cmdstr = ' '.join(shlex.quote(str(arg)) for arg in args)
            if trailing_space:
                cmdstr += ' '
            self.tui.widgets.show('cli')
            self.tui.widgets.cli.base_widget.set_edit_text(cmdstr)
            self.tui.widgets.cli.base_widget.set_edit_pos(len(cmdstr))


class InteractiveCmd(mixin.placeholders, metaclass=InitCommand):
    name = 'interactive'
    provides = {'tui'}
    category = 'tui'
    description = 'Complete partial command with user input from a dialog'
    usage = ('interactive <COMMAND> [<OPTIONS>]',)
    examples = (
        'interactive "move \'[{location}/]\'"',
        '\tAsk for the destination directory when moving torrents.',
        '',
        'tab ls & interactive "limit \'[]\'" --per-change --on-cancel "tab --close --focus left"',
        ('\tOpen a new tab with all torrents and filter them as you type.  '
         'Keep the tab open if the user input field is accepted with <enter> '
         'or close the tab and focus the previous one if the dialog is aborted '
         'with <escape>.'),
        '',
        'tab ls stopped & interactive \'limit "[]"\' -p -a "mark --all & start" -x "tab --close --focus left"',
        ('\tSearch for stopped torrents only.  When accepted, the matching torrents '
         'are started.  The new tab is always closed, whether the dialog is '
         'accepted or not.'),
    )
    argspecs = (
        { 'names': ('COMMAND',),
          'description': ('Any command with "[PREFILLED TEXT]" as marker for '
                          'user input field (see USER INPUT FIELDS) and '
                          '"{{NAME}}" as placeholder for values of the currently '
                          'focused list item (see PLACEHOLDERS)') },
        { 'names': ('--per-change', '-p'), 'action': 'store_true',
          'description': 'Whether to run COMMAND every time the input is changed'},
        { 'names': ('--on-accept', '-a'), 'metavar': 'ACCEPT COMMAND',
          'description': 'Command to run when the dialog is accepted (with <enter>)' },
        { 'names': ('--on-cancel', '-c'), 'metavar': 'CANCEL COMMAND',
          'description': 'Command to run when the dialog is aborted (with <escape>)' },
        { 'names': ('--on-close', '-x'), 'metavar': 'CLOSE COMMAND',
          'description': 'Command to run after the dialog is closed either way' },
        { 'names': ('--ignore-errors', '-i'), 'action': 'store_true',
          'description': 'Whether to ignore errors from COMMAND' },
    )
    more_sections = {
        'COMMANDS': (('For each occurence of "[]" in any command, the user is '
                      'prompted for input to insert at that point.  Any text between '
                      '"[" and "]" is used as the initial user input.  "[" can be '
                      'escaped with "\\" in which case the corresponding "]" is also '
                      'interpreted literally.'),
                     '',
                     ('COMMAND is called if the user presses <enter> or, if --per-change '
                      'is given, after any user input field is changed.'),
                     '',
                     ('COMMAND must contain at least one user input field.  Any of the '
                      'commands described below are called without user interaction if '
                      'they don\'t contain any user input fields.'),
                     '',
                     ('ACCEPT COMMAND is called after COMMAND if the user accepts the '
                      'dialog by pressing <enter>.'),
                     '',
                     'CANCEL COMMAND is called if the user aborts the COMMAND dialog.',
                     '',
                     ('CLOSE COMMAND is always called when the dialog is closed either '
                      'by accepting or by cancelling it.')),
        'PLACEHOLDERS': mixin.placeholders.HELP,
    }

    tui = ExpectedResource
    cmdmgr = ExpectedResource
    cfg = ExpectedResource

    import re
    _input_regex = re.compile(r'(?<!\\)(\[.*?\])')

    async def run(self, COMMAND, per_change, on_accept, on_cancel, on_close, ignore_errors):
        cmd = await self._parse_cmd(COMMAND)
        accept_cmd = await self._parse_cmd(on_accept) if on_accept else None
        cancel_cmd = await self._parse_cmd(on_cancel) if on_cancel else None
        close_cmd = await self._parse_cmd(on_close) if on_close else None
        self._ignore_errors = ignore_errors

        if len(cmd) == 1:
            # There are no user input markers
            raise CmdError('No user input fields ("[]"): %s' % COMMAND)

        def close_cb():
            self._run_cmd_or_open_dialog(close_cmd)

        if per_change:
            def accept_cb():
                self._run_cmd_from_dialog()
                self._run_cmd_or_open_dialog(accept_cmd)

            def cancel_cb():
                self._run_cmd_or_open_dialog(cancel_cmd)

            self._open_dialog(cmd,
                              on_change=self._run_cmd_from_dialog,
                              on_accept=accept_cb,
                              on_cancel=cancel_cb,
                              on_close=close_cb)
        else:
            def accept_cb():
                self._run_cmd_from_dialog()
                self._run_cmd_or_open_dialog(accept_cmd)

            def cancel_cb():
                self._run_cmd_or_open_dialog(cancel_cmd)

            self._open_dialog(cmd,
                              on_accept=accept_cb,
                              on_cancel=cancel_cb,
                              on_close=close_cb)

    _WIDGET_NAME = 'interactive_prompt'
    _MIN_EDIT_WIDTH = 25
    _MAX_EDIT_WIDTH = 50
    def _open_dialog(self, cmd, on_change=None, on_accept=None, on_cancel=None, on_close=None):
        import urwid
        from ...tui.cli import CLIEditWidget

        def accept_cb(widget):
            # CLIEditWidget only automatically appends to history when it gets
            # an <enter> key, but only one gets it if there are multiple user
            # input fields.
            for part in self._edit_widgets:
                part.append_to_history()
            self._close_dialog()
            if on_accept: on_accept()
            if on_close: on_close()

        def cancel_cb(widget):
            self._close_dialog()
            if on_cancel: on_cancel()
            if on_close: on_close()

        def change_cb(widget):
            if on_change: on_change()

        # Derive history file name from command
        import re
        filename = re.sub('[/\n]', '__', ''.join(cmd))
        history_file_base = os.path.join(self.cfg['tui.cli.history-dir'], filename)

        columns_args = [('pack', urwid.Text(':'))]
        self._cmd_parts = []
        self._edit_widgets = []
        edit_index = 0
        for part in cmd:
            if part[0] == '[' and part[-1] == ']':
                edit_index += 1
                history_file = history_file_base + '.input%d' % edit_index
                log.debug('History file for edit #%d: %r', edit_index, history_file)
                edit_widget = CLIEditWidget(on_change=change_cb,
                                            on_accept=accept_cb,
                                            on_cancel=cancel_cb,
                                            history_file=history_file)
                edit_widget.set_edit_text(part[1:-1])
                edit_widget.set_edit_pos(len(part))
                columns_args.append(urwid.AttrMap(edit_widget, 'prompt'))
                self._cmd_parts.append(edit_widget)
                self._edit_widgets.append(edit_widget)
            else:
                columns_args.append(('pack', urwid.Text(part)))
                self._cmd_parts.append(part)

        class MyColumns(urwid.Columns):
            """Use <tab> and <shift-tab> to move focus between input fields"""
            def keypress(self, size, key):
                def move_right():
                    if self.focus_position < len(self.contents)-1:
                        self.focus_position += 1
                    else:
                        self.focus_position = 0

                def move_left():
                    if self.focus_position > 0:
                        self.focus_position -= 1
                    else:
                        self.focus_position = len(self.contents)-1

                if key == 'tab':
                    move_right()
                    while not isinstance(self.focus.base_widget, urwid.Edit):
                        move_right()
                elif key == 'shift-tab':
                    move_left()
                    while not isinstance(self.focus.base_widget, urwid.Edit):
                        move_left()
                else:
                    log.debug('focus pos: %r', self.focus_position)
                    return super().keypress(size, key)

        columns_widget = MyColumns(columns_args)

        # Close any previously opened dialog
        if self.tui.widgets.exists(self._WIDGET_NAME):
            self._close_dialog()

        # Focus the first empty input widget if there are any
        for i,(w,_) in enumerate(columns_widget.contents):
            w = w.base_widget
            log.debug('%02d: %r', i, w)
            if hasattr(w, 'edit_text') and w.edit_text == '':
                columns_widget.focus_position = i
                break

        self.tui.widgets.add(name=self._WIDGET_NAME,
                             widget=urwid.AttrMap(columns_widget, 'cli'),
                             position=self.tui.widgets.get_position('cli'),
                             removable=True,
                             options='pack')

    def _close_dialog(self):
        self.tui.widgets.remove(self._WIDGET_NAME)
        self.tui.widgets.focus_name = 'main'

    def _run_cmd_or_open_dialog(self, cmd):
        if not cmd:
            return
        elif len(cmd) == 1:
            log.debug('Running command without dialog: %r', cmd)
            self._run_cmd(cmd[0])
        else:
            log.debug('Running command in dialog: %r', cmd)
            self._open_dialog(cmd, on_accept=self._run_cmd_from_dialog)

    def _run_cmd_from_dialog(self):
        cmd = []
        for part in self._cmd_parts:
            if hasattr(part, 'edit_text'):
                cmd.append(part.edit_text)
            else:
                cmd.append(part)
        cmd = ''.join(cmd)
        log.debug('Got command from current dialog: %r', cmd)
        self._run_cmd(cmd)

    def _run_cmd(self, cmd):
        log.debug('Running cmd: %r', cmd)
        if self._ignore_errors:
            # Overload the error() method on the command's instance
            self.cmdmgr.run_task(cmd, error=lambda msg: None)
        else:
            self.cmdmgr.run_task(cmd)

    async def _parse_cmd(self, cmd):
        assert isinstance(cmd, str)
        args = await self.parse_placeholders(cmd)
        return self._split_cmd_at_inputs(args[0])

    def _split_cmd_at_inputs(self, cmd):
        """
        Split `cmd` so that each input marker ("[...]") is a single item

        Example result:
            ['somecmd --an-argument ', '[user input goes here]', ' some more arguments']
        """

        log.debug('Splitting %r', cmd)
        parts = [part for part in self._input_regex.split(cmd) if part]
        log.debug('Split: %r', parts)

        for i in range(len(parts)):
            parts[i] = parts[i].replace('\\[', '[')
        log.debug('Unescaped: %r', parts)

        return parts


class MarkCmd(metaclass=InitCommand):
    name = 'mark'
    provides = {'tui'}
    category = 'tui'
    description = 'Select torrents or files for an action'
    usage = ('mark [<OPTIONS>]',)
    argspecs = (
        { 'names': ('--focus-next','-n'), 'action': 'store_true',
          'description': 'Move focus forward after marking or toggling' },
        { 'names': ('--toggle','-t'), 'action': 'store_true',
          'description': 'Mark if unmarked, unmark if marked' },
        { 'names': ('--all','-a'), 'action': 'store_true',
          'description': 'Mark or toggle all items' },
    )
    more_sections = {
        'NOTES': (('The column "marked" must be in the "columns.*" settings. Otherwise '
                   'marked list items are indistinguishable from unmarked ones.'),
                  '',
                  ('The character that is displayed in the "marked" column is '
                   'specified by the settings "tui.marked.on" and "tui.marked.off".')),
    }

    tui = ExpectedResource

    def run(self, focus_next, toggle, all):
        widget = self.tui.tabs.focus
        if not widget.has_marked_column:
            raise CmdError('Nothing to mark here.')
        else:
            widget.mark(toggle=toggle, all=all)
            if focus_next:
                widget.focus_position += 1


class UnmarkCmd(metaclass=InitCommand):
    name = 'unmark'
    provides = {'tui'}
    category = 'tui'
    description = 'Deselect torrents or files for an action'
    usage = ('unmark [<OPTIONS>]',)
    argspecs = (
        { 'names': ('--focus-next','-n'), 'action': 'store_true',
          'description': 'Move focus forward after unmarking or toggling' },
        { 'names': ('--toggle','-t'), 'action': 'store_true',
          'description': 'Mark if unmarked, unmark if marked' },
        { 'names': ('--all','-a'), 'action': 'store_true',
          'description': 'Unmark or toggle all items' },
    )
    more_sections = MarkCmd.more_sections

    tui = ExpectedResource

    def run(self, focus_next, toggle, all):
        widget = self.tui.tabs.focus
        if not widget.has_marked_column:
            raise CmdError('Nothing to unmark here.')
        else:
            widget.unmark(toggle=toggle, all=all)
            if focus_next:
                widget.focus_position += 1


class QuitCmd(metaclass=InitCommand):
    name = 'quit'
    provides = {'tui'}
    category = 'tui'
    description = 'Terminate the TUI'

    def run(self):
        import urwid
        raise urwid.ExitMainLoop()


class FindCmd(metaclass=InitCommand):
    name = 'find'
    provides = {'tui'}
    category = 'tui'
    description = 'Find text in the content of the focused tab'
    usage = ('find [<OPTIONS>] [<PHRASE>]',)
    argspecs = (
        { 'names': ('--clear','-c'), 'action': 'store_true',
          'description': ('Remove previously applied filter; this is '
                          'the default if no PHRASE arguments are provided') },
        { 'names': ('--next','-n'), 'action': 'store_true',
          'description': 'Jump to next match (call `find <PHRASE>` first)' },
        { 'names': ('--previous','-p'), 'action': 'store_true',
          'description': 'Jump to previous match (call `find <PHRASE>` first)' },
        { 'names': ('PHRASE',), 'nargs': '*',
          'description': 'Search phrase' },
    )
    tui = ExpectedResource

    def run(self, clear, next, previous, PHRASE):
        content = self.tui.tabs.focus.base_widget
        if not hasattr(content, 'search_phrase'):
            raise CmdError('This tab does not support finding.')
        elif next and previous:
            raise CmdError('The options --next and --previous contradict each other.')
        elif next:
            if content.search_phrase is None:
                raise CmdError('Set a search phrase first with `find <PHRASE>`.')
            else:
                content.jump_to_next_match()
        elif previous:
            if content.search_phrase is None:
                raise CmdError('Set a search phrase first with `find <PHRASE>`.')
            else:
                content.jump_to_prev_match()
        elif clear:
            content.search_phrase = None
        else:
            try:
                content.search_phrase = ' '.join(PHRASE)
                content.maybe_jump_to_next_match()
            except ValueError as e:
                raise CmdError(e)


class LimitCmd(metaclass=InitCommand):
    name = 'limit'
    provides = {'tui'}
    category = 'tui'
    description = 'Limit contents of the focused tab by applying more filters'
    usage = ('limit [<OPTIONS>] [<FILTER> <FILTER> ...]',)
    argspecs = (
        { 'names': ('--clear','-c'), 'action': 'store_true',
          'description': ('Remove previously applied filter; this is '
                          'the default if no FILTER arguments are provided') },
        { 'names': ('FILTER',), 'nargs': '?',
          'description': 'Filter expression (see `help filter`)' },
    )
    tui = ExpectedResource

    def run(self, clear, FILTER):
        content = self.tui.tabs.focus.base_widget
        if not hasattr(content, 'secondary_filter'):
            raise CmdError('This tab does not support limiting.')
        else:
            if clear or not FILTER:
                content.secondary_filter = None
            else:
                try:
                    content.secondary_filter = FILTER
                except ValueError as e:
                    raise CmdError(e)


class SortCmd(metaclass=InitCommand):
    name = 'sort'
    aliases = ()
    provides = {'tui'}
    category = 'tui'
    description = "Sort lists of torrents/peers/trackers/etc"
    usage = ('sort [<OPTIONS>] [<ORDER> <ORDER> <ORDER> ...]',)
    examples = ('sort tracker status !rate-down',
                'sort --add eta')
    argspecs = (
        {'names': ('ORDER',), 'nargs': '*',
         'description': 'How to sort list items (see SORT ORDERS section)'},

        {'names': ('--add', '-a'), 'action': 'store_true',
         'description': 'Append ORDERs to current list of sort orders instead of replacing it'},

        {'names': ('--reset', '-r'), 'action': 'store_true',
         'description': 'Go back to sort order that was used when list was created'},

        {'names': ('--none', '-n'), 'action': 'store_true',
         'description': 'Remove all sort orders from the list'},
    )

    def _list_sort_orders(title, sortercls):
        return (title,) + \
            tuple('\t{}\t - \t{}'.format(', '.join((sname,) + s.aliases), s.description)
                  for sname,s in sorted(sortercls.SORTSPECS.items()))

    from ...client import (TorrentSorter, TorrentPeerSorter,
                           TorrentTrackerSorter, SettingSorter)
    more_sections = {
        'SORT ORDERS': _list_sort_orders('TORRENT LISTS', TorrentSorter) +
                       ('',) +
                       _list_sort_orders('PEER LISTS', TorrentPeerSorter) +
                       ('',) +
                       _list_sort_orders('TRACKER LISTS', TorrentTrackerSorter)
    }

    tui = ExpectedResource

    async def run(self, add, reset, none, ORDER):
        current_tab = self.tui.tabs.focus

        if reset:
            current_tab.sort = 'RESET'

        if none:
            current_tab.sort = None

        if ORDER:
            # Find appropriate sorter class for focused list
            from ...tui.views.torrent_list import TorrentListWidget
            from ...tui.views.peer_list import PeerListWidget
            from ...tui.views.tracker_list import TrackerListWidget
            from ...tui.views.setting_list import SettingListWidget
            if isinstance(current_tab, TorrentListWidget):
                sortcls = self.TorrentSorter
            elif isinstance(current_tab, PeerListWidget):
                sortcls = self.TorrentPeerSorter
            elif isinstance(current_tab, TrackerListWidget):
                sortcls = self.TorrentTrackerSorter
            elif isinstance(current_tab, SettingListWidget):
                sortcls = self.SettingSorter
            else:
                raise CmdError('Current tab is not sortable.')

            try:
                new_sort = sortcls(ORDER)
            except ValueError as e:
                raise CmdError(e)

            if add and current_tab.sort is not None:
                current_tab.sort += new_sort
            else:
                current_tab.sort = new_sort


class TabCmd(mixin.select_torrents, metaclass=InitCommand):
    name = 'tab'
    provides = {'tui'}
    category = 'tui'
    description = 'Open, close and focus tabs'
    usage = ('tab [<OPTIONS>]',
             'tab [<OPTIONS>] <COMMAND>')
    examples = ('tab',
                'tab -c',
                'tab -c active',
                'tab ls active',
                'tab -b ls active',
                'tab -f active',
                'tab -f 3 ls active',
                'tab -b -f -1 ls active')
    argspecs = (
        { 'names': ('--background', '-b'), 'action': 'store_true',
          'description': 'Do not focus new tab' },
        { 'names': ('--close-all', '-C'), 'action': 'store_true',
          'description': 'Close all tabs' },
        { 'names': ('--close', '-c'), 'nargs': '?', 'default': False, 'document_default': False,
          'description': 'Close focused or specified tab (see TAB IDENTIFIERS SECTION)' },
        { 'names': ('--focus', '-f'),
          'description': 'Focus specified tab (see TAB IDENTIFIERS SECTION)' },
        { 'names': ('--title', '-t'),
          'description': 'Manually set tab title instead of generating one' },
        { 'names': ('COMMAND',), 'nargs': 'REMAINDER',
          'description': ('Command to run in new tab') },
    )
    more_sections = {
        'TAB IDENTIFIERS': (
            'There are three ways to specify a tab (e.g. to close it):',
            ('  - \tIntegers specify the position of the tab.  Positive numbers '
             'start from the left and negative numbers start from the right '
             '(1 (and 0) is the leftmost tab and -1 is the rightmost tab).'),
            ('  - \t"left" and "right" specify the tabs right and left to the '
             'currently focused tab.'),
            ('  - \tAnything else is assumed to be a part of a tab title.  If there '
             'are multiple matches, the first match from the left wins.'),
        ),
    }

    tui = ExpectedResource
    cmdmgr = ExpectedResource

    async def run(self, close, close_all, focus, background, title, COMMAND):
        tabs = self.tui.tabs

        # Find relevant tab IDs and fail immediately if unsuccessful
        tabid_old = tabs.get_id()
        if focus is not None:
            tabid_focus = self._get_tab_id(focus)
            if tabid_focus is None:
                raise CmdError('No such tab: %r' % (focus,))
        if close is not False:
            tabid_close = self._get_tab_id(close)
            if tabid_close is None:
                raise CmdError('No such tab: %r' % (close,))

        # COMMAND may get additional hidden arguments as instance attributes
        cmd_attrs = {}

        # The command we're running might be interested in the items the user
        # had selected in the previously focused tab, e.g. if the user selects
        # multiple torrents and runs "tab filelist", the new tab should list the
        # files of the selected torrnets.  So we provide the ID of the
        # previously focused tab as a command attribute.  The command can then
        # use that to look up the relevant torrents (usually via the
        # select_torrents() mixin class).
        if tabid_old is not None:
            cmd_attrs['previous_tab_id'] = tabid_old

        # Apply close/focus operations
        if focus is not None:
            log.debug('Focusing tab %r', tabid_focus)
            tabs.focus_id = tabid_focus
        if close_all is not False:
            log.debug('Closing all tabs')
            tabs.clear()
        elif close is not False:
            log.debug('Closing tab %r', tabid_close)
            tabs.remove(tabid_close)

        # If no tabs were closed or focused, open a new one
        if close is False and close_all is False and focus is None:
            titlew = make_tab_title_widget(title or 'Empty tab',
                                           attr_unfocused='tabs.unfocused',
                                           attr_focused='tabs.focused')
            tabs.insert(titlew, position='right')
            log.debug('Inserted new tab at position %d: %r', tabs.focus_position, titlew.base_widget.text)

        # Maybe provide a user-specified tab title to the new command
        if title:
            cmd_attrs['title'] = title

        if COMMAND:
            # Execute command
            cmd_str = ' '.join(shlex.quote(arg) for arg in COMMAND)
            log.debug('Running command in tab %s with args %s: %r',
                      tabs.focus_position,
                      ', '.join('%s=%r' % (k,v) for k,v in cmd_attrs.items()),
                      cmd_str)

            success = await self.cmdmgr.run_async(cmd_str, **cmd_attrs)
        else:
            success = True

        if background:
            tabs.focus_id = tabid_old

        return success

    def _get_tab_id(self, pos):
        tabs = self.tui.tabs
        if len(tabs) == 0:
            return None

        if pos is None:
            return tabs.focus_id

        def find_id_by_index(index):
            try:
                index = int(index)
            except ValueError:
                pass
            else:
                index_max = len(tabs) - 1
                # Internally, first tab is at index 0, but for users it's 1, unless
                # they gave us 0, in which case we assume they mean 1.
                index = index-1 if index > 0 else index

                # Limit index to index_max, considering negative values when
                # indexing from the right.
                if index < 0:
                    index = max(index, -index_max-1)
                else:
                    index = min(index, index_max)
                return tabs.get_id(index)

        def find_right_left_id(right_or_left):
            tabcount = len(tabs)
            if tabcount > 1:
                cur_index = tabs.focus_position
                cur_index = 1 if cur_index is None else cur_index
                if right_or_left == 'left':
                    return tabs.get_id(max(0, cur_index-1))
                elif right_or_left == 'right':
                    return tabs.get_id(min(tabcount-1, cur_index+1))

        def find_id_by_title(string):
            for index,title in enumerate(tabs.titles):
                if string in title.original_widget.text:
                    return tabs.get_id(index)

        # Try to use pos as an index
        tabid = find_id_by_index(pos)
        if tabid is not None:
            log.debug('Found tab ID by index: %r -> %r', pos, tabid)
            return tabid

        pos_str = str(pos)

        # Move to left/right tab
        tabid = find_right_left_id(pos_str)
        if tabid is not None:
            log.debug('Found tab ID by direction: %r -> %r', pos, tabid)
            return tabid

        # Try to find tab title
        tabid = find_id_by_title(pos_str)
        if tabid is not None:
            log.debug('Found tab ID by title: %r -> %r', pos, tabid)
            return tabid


class TUICmd(metaclass=InitCommand):
    name = 'tui'
    provides = {'tui'}
    category = 'tui'
    description = 'Show or hide parts of the text user interface'
    usage = ('tui <ACTION> <ELEMENT> <ELEMENT> ...',)
    examples = ('tui toggle log',
                'tui hide topbar.help')
    argspecs = (
        { 'names': ('ACTION',), 'choices': ('show', 'hide', 'toggle'),
          'description': '"show", "hide" or "toggle"' },
        { 'names': ('ELEMENT',), 'nargs': '+',
          'description': ('Name of TUI elements; '
                          'see ELEMENT NAMES section for a list') },
    )

    # Lazily load the element names from tui (HelpManager supports sequences
    # of lines or a callable that returns them)
    def __create_element_names():
        from ...tui.main import widgets
        return ('Available TUI element names are: ' +
                ', '.join(str(e) for e in sorted(widgets.names_recursive)),)
    more_sections = { 'ELEMENT NAMES': __create_element_names }

    tui = ExpectedResource

    def run(self, ACTION, ELEMENT):
        widgets = self.tui.widgets
        widget = None
        success = True
        for element in ELEMENT:
            # Resolve path
            path = element.split('.')
            target_name = path.pop(-1)
            current_path = []
            widget = widgets
            try:
                for widgetname in path:
                    current_path.append(widgetname)
                    widget = getattr(widget, widgetname)
            except AttributeError as e:
                self.error('Unknown TUI element: %r' % ('.'.join(current_path),))

            if widget is not None:
                action = getattr(widget, ACTION)
                if any(ACTION == x for x in ('hide', 'toggle')):
                    action = partial(action, free_space=False)

                log.debug('%sing %s in %s', ACTION.capitalize(), target_name, widget)
                try:
                    action(target_name)
                except ValueError as e:
                    success = False
                    self.error(e)
                else:
                    success = success and True

        if not success:
            raise CmdError()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Column classes that display information in peer lists"""

import urwid

from ..table import ColumnHeaderWidget
from . import (Style, CellWidgetBase)
from ...views.tracker import COLUMNS as _COLUMNS


TUICOLUMNS = {}

class Torrent(_COLUMNS['torrent'], CellWidgetBase):
    width = ('weight', 50)
    style = Style(prefix='trackerlist.torrent', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['torrent'].header),
                           style.attrs('header'))

TUICOLUMNS['torrent'] = Torrent


class Tier(_COLUMNS['tier'], CellWidgetBase):
    style = Style(prefix='trackerlist.tier', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['tier'].header),
                           style.attrs('header'))

TUICOLUMNS['tier'] = Tier


class Domain(_COLUMNS['domain'], CellWidgetBase):
    width = ('weight', 20)
    style = Style(prefix='trackerlist.domain', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['domain'].header),
                           style.attrs('header'))

TUICOLUMNS['domain'] = Domain


class AnnounceURL(_COLUMNS['url-announce'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='trackerlist.url-announce', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['url-announce'].header),
                           style.attrs('header'))

TUICOLUMNS['url-announce'] = AnnounceURL


class ScrapeURL(_COLUMNS['url-scrape'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='trackerlist.url-scrape', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['url-scrape'].header),
                           style.attrs('header'))

TUICOLUMNS['url-scrape'] = ScrapeURL


class Status(_COLUMNS['status'], CellWidgetBase):
    style = Style(prefix='trackerlist.status', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['status'].header),
                           style.attrs('header'))

TUICOLUMNS['status'] = Status


class Error(_COLUMNS['error'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='trackerlist.error', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['error'].header),
                           style.attrs('header'))

TUICOLUMNS['error'] = Error


class ErrorAnnounce(_COLUMNS['error-announce'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='trackerlist.error-announce', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['error-announce'].header),
                           style.attrs('header'))

TUICOLUMNS['error-announce'] = ErrorAnnounce


class ErrorScrape(_COLUMNS['error-scrape'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='trackerlist.error-scrape', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['error-scrape'].header),
                           style.attrs('header'))

TUICOLUMNS['error-scrape'] = ErrorScrape


class Downloads(_COLUMNS['downloads'], CellWidgetBase):
    style = Style(prefix='trackerlist.downloads', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['downloads'].header),
                           style.attrs('header'))

TUICOLUMNS['downloads'] = Downloads


class Leeches(_COLUMNS['leeches'], CellWidgetBase):
    style = Style(prefix='trackerlist.leeches', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['leeches'].header),
                           style.attrs('header'))

TUICOLUMNS['leeches'] = Leeches


class Seeds(_COLUMNS['seeds'], CellWidgetBase):
    style = Style(prefix='trackerlist.seeds', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['seeds'].header),
                           style.attrs('header'))

TUICOLUMNS['seeds'] = Seeds


class LastAnnounce(_COLUMNS['last-announce'], CellWidgetBase):
    style = Style(prefix='trackerlist.last-announce', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['last-announce'].header),
                           style.attrs('header'))

TUICOLUMNS['last-announce'] = LastAnnounce


class NextAnnounce(_COLUMNS['next-announce'], CellWidgetBase):
    style = Style(prefix='trackerlist.next-announce', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['next-announce'].header),
                           style.attrs('header'))

TUICOLUMNS['next-announce'] = NextAnnounce


class LastScrape(_COLUMNS['last-scrape'], CellWidgetBase):
    style = Style(prefix='trackerlist.last-scrape', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['last-scrape'].header),
                           style.attrs('header'))

TUICOLUMNS['last-scrape'] = LastScrape


class NextScrape(_COLUMNS['next-scrape'], CellWidgetBase):
    style = Style(prefix='trackerlist.next-scrape', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['next-scrape'].header),
                           style.attrs('header'))

TUICOLUMNS['next-scrape'] = NextScrape



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from .. import (InitCommand, CmdError, ExpectedResource, utils)
from ._common import (make_X_FILTER_spec, make_COLUMNS_doc,
                      make_SORT_ORDERS_doc, make_SCRIPTING_doc)
from ...utils.usertypes import Float
from . import _mixin as mixin

import subprocess
import operator


class RcCmdbase(metaclass=InitCommand):
    name = 'rc'
    aliases = ('source',)
    category = 'configuration'
    provides = set()
    description = 'Run commands in rc file'
    usage = ('rc <FILE>',)
    examples = ('rc rc.example   # Load $XDG_CONFIG_HOME/.config/{__appname__}/rc.example',)
    argspecs = (
        {'names': ('FILE',),
         'description': ('Path to rc file; if FILE does not start with '
                         "'.', '~' or '/', $XDG_CONFIG_HOME/.config/{__appname__}/ "
                         'is prepended')},
    )
    cmdmgr = ExpectedResource

    async def run(self, FILE):
        from ...settings import rcfile
        from ...settings.defaults import DEFAULT_RCFILE
        import os

        if not os.path.isabs(FILE) and not os.path.exists(FILE):
            default_dir = os.path.dirname(DEFAULT_RCFILE)
            FILE = os.path.join(default_dir, FILE)

        try:
            lines = rcfile.read(FILE)
        except rcfile.RcFileError as e:
            raise CmdError('Loading rc file failed: %s' % e)
        else:
            log.debug('Running commands from rc file: %r', FILE)
            for cmdline in lines:
                success = await self.cmdmgr.run_async(cmdline)
                # False means failure, None means the command didn't run because
                # the active interface doesn't support it
                if success is False:
                    raise CmdError()


class ResetCmdbase(metaclass=InitCommand):
    name = 'reset'
    category = 'configuration'
    provides = set()
    description = 'Reset settings to their default values'
    usage = ('reset <NAME> <NAME> <NAME> ...',)
    examples = ('reset connect.port',)
    argspecs = (
        {'names': ('NAME',), 'nargs': '*',
         'description': 'Name of setting'},
    )
    more_sections = {
        'SEE ALSO': ('Run `help settings` for a list of all available settings.',
                     'Note that remote settings (srv.*) cannot be reset.'),
    }
    cfg = ExpectedResource
    srvcfg = ExpectedResource

    def run(self, NAME):
        if not NAME:
            raise CmdError('Missing NAME argument')

        success = True
        for name in NAME:
            if name.startswith('srv.') and name[4:] in self.srvcfg:
                self.error('Remote settings cannot be reset: %s' % name)
                success = False
            elif name not in self.cfg:
                self.error('Unknown setting: %s' % name)
                success = False
            else:
                self.cfg.reset(name)
        if not success:
            raise CmdError()


from ...client import ClientError
class SetCmdbase(mixin.get_setting_sorter, mixin.get_setting_columns,
                 metaclass=InitCommand):
    name = 'set'
    category = 'configuration'
    provides = set()
    description = 'Change {__appname__} settings'
    usage = ('set [<NAME>[:eval]] [<VALUE>]',)
    examples = ('set connect.host my.server.example.org',
                'set connect.user jonny_sixpack',
                'set connect.password:eval getpw --id transmission',
                'set tui.log.height +=10')
    argspecs = (
        {'names': ('NAME',), 'nargs': '?',
         'description': "Name of setting; append ':eval' to turn VALUE into a shell command"},

        {'names': ('VALUE',), 'nargs': 'REMAINDER',
         'description': ('New value or shell command that prints the new value to stdout; '
                         "numerical values can be adjusted by prepending '+=' or '-='")},

        { 'names': ('--sort', '-s'),
          'description': 'Comma-separated list of sort orders (see SORT ORDERS section)' },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'columns.settings' setting",
          'description': 'Comma-separated list of column names (see COLUMNS section)' },
    )
    from ...views.setting import COLUMNS
    from ...client.sorters.setting import SettingSorter
    more_sections = {
        'COLUMNS': make_COLUMNS_doc(COLUMNS, '--columns', 'columns.settings'),
        'SORT ORDERS': make_SORT_ORDERS_doc(SettingSorter, '--sort', 'sort.settings'),
        'SCRIPTING': make_SCRIPTING_doc(name),
        'SEE ALSO': (('Run `help settings` or run `set` without commands '
                      'for a list of available local and remote settings.'),),
    }
    cfg = ExpectedResource
    srvcfg = ExpectedResource

    async def run(self, NAME, VALUE, sort, columns):
        if not NAME and not VALUE:
            # Get remote setting values
            try:
                await self.srvcfg.update()
            except ClientError as e:
                error = e
            else:
                error = None

            # Show list of settings
            sort = self.cfg['sort.settings'] if sort is None else sort
            columns = self.cfg['columns.settings'] if columns is None else columns
            try:
                sort = self.get_setting_sorter(sort)
                columns = self.get_setting_columns(columns)
            except ValueError as e:
                raise CmdError(e)
            else:
                self.make_setting_list(sort, columns)
                if error:
                    raise CmdError(error)
            return

        # NAME might have ':eval' attached if VALUE is shell command
        # `name` is the user-facing name of the variable.
        # `key` is the lookup key in the config mapping.
        try:
            name, key, is_local_setting = self._parse_name(NAME)
        except ValueError as e:
            raise CmdError(e)

        cfg = self.cfg if is_local_setting else self.srvcfg

        # Get current value in case we want to adjust it
        if not is_local_setting:
            try:
                await cfg.update()
            except ClientError as e:
                raise CmdError(e)

        # VALUE might be shell command or have '+='/'-=' prepended
        try:
            value = self._parse_value(VALUE,
                                      listify=isinstance(cfg[key], tuple),
                                      is_cmd=NAME.endswith(':eval'))
        except ValueError as e:
            # Report potential stderr output if VALUE is a command
            raise CmdError('%s: %s' % (name, e))

        # Separate '+=' or '-=' from value
        try:
            op, value = self._get_operator(value)
        except ValueError as e:
            # Report invalid value after operator (e.g. nan)
            raise CmdError('%s = %s: %s' % (name, self._stringify(value), e))

        # Value may have an operator (e.g. '+=' or '-=') to adjust the current value
        if op is not None:
            try:
                value = self._adjust_value(cfg[key], op, value)
            except ValueError as e:
                # Report out-of-bounds value
                opfunc = getattr(operator, op)
                unbound = cfg[key].copy(min=-float('inf'), max=float('inf'))
                invalid = opfunc(unbound, value)
                raise CmdError('%s = %s: %s' % (name, self._stringify(invalid), e))

        # Update setting's value
        try:
            if is_local_setting:
                log.debug('Local setting: %r = %r', name, value)
                cfg[key] = value
            else:
                log.debug('Remote setting: %r = %r', name, value)
                await cfg.set(key, value)
        except ValueError as e:
            raise CmdError('%s = %s: %s' % (name, self._stringify(value), e))
        except ClientError as e:
            raise CmdError(e)

    def _parse_name(self, name):
        # Return (name, key, is_local_setting)
        if name.endswith(':eval'):
            name = name[:-5]
        if name in self.cfg:
            return name, name, True
        elif name.startswith('srv.') and name[4:] in self.srvcfg:
            return name, name[4:], False
        else:
            raise ValueError('Unknown setting: %s' % name)

    def _parse_value(self, value, listify=False, is_cmd=False):
        if is_cmd:
            value = [self._eval_cmd(value)]
        if listify:
            return utils.listify_args(value)
        else:
            return ' '.join(value)

    @staticmethod
    def _eval_cmd(cmd):
        if not isinstance(cmd, str):
            cmd = ' '.join(cmd)
        log.debug('Running shell command: %r', cmd)
        proc = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout = proc.stdout.decode('utf-8').strip('\n')
        stderr = proc.stderr.decode('utf-8').strip('\n')
        if stderr:
            raise ValueError(stderr)
        else:
            return stdout

    @staticmethod
    def _get_operator(value):
        if isinstance(value, str):
            value = value.strip()
            if len(value) >= 3:
                def to_num(string):
                    try:
                        return Float(string)
                    except ValueError as e:
                        raise ValueError('%s: %r' % (e, string))
                if value[:2] == '+=':
                    return '__add__', to_num(value[2:])
                elif value[:2] == '-=':
                    return '__sub__', to_num(value[2:])
        return None, value

    @staticmethod
    def _adjust_value(current, op, value):
        if isinstance(current, (float, int)):
            if current >= float('inf'):
                current = Float(0)
            func = getattr(operator, op)
            value = func(current, value)
        return value

    @staticmethod
    def _stringify(value):
        from collections.abc import Iterable
        if not isinstance(value, str) and isinstance(value, Iterable):
            return ', '.join(str(item) for item in value)
        else:
            return str(value)


class RateLimitCmdbase(metaclass=InitCommand):
    name = 'ratelimit'
    aliases = ('rate', 'rl')
    provides = set()
    category = 'configuration'
    description = 'Limit transfer rates per torrent or globally'
    usage = ('ratelimit',
             'ratelimit <DIRECTION>',
             'ratelimit <DIRECTION> <LIMIT>',
             'ratelimit <DIRECTION> <LIMIT> <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('ratelimit up 5Mb',
                'ratelimit down,up 1M global',
                'ratelimit up,dn off "This torrent" size<100MB')
    argspecs = (
        {'names': ('DIRECTION',), 'nargs': '?', 'default': 'up,down',
         'description': 'Any combination of "up", "down" or "dn" separated by a comma'},

        {'names': ('LIMIT',), 'nargs': '?',
         'description': ('Maximum allowed transfer rate (see `help srv.limit.rate.up` '
                         'for the syntax) or "show" to display the current limit')},

        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*',
                           more_text=('"global" to set global limit (same as setting '
                                      "srv.limit.rate.<DIRECTION>); may be omitted in CLI mode "
                                      'for the same effect as specifying "global"')),

        { 'names': ('--quiet','-q'), 'action': 'store_true',
          'description': 'Do not show new bandwidth rate(s)' },
    )
    srvapi = ExpectedResource

    async def run(self, DIRECTION, LIMIT, TORRENT_FILTER, quiet):
        directions = tuple('down' if d == 'dn' else d
                           for d in map(str.lower, DIRECTION.split(',')))
        for d in directions:
            if d not in ('up', 'down'):
                raise CmdError('Invalid direction: %r' % (d,))

        # _show_limits() and _set_limits() are defined in cli.config and
        # tui.config because the TUI can use the focused torrent while the CLI
        # can't.
        if not LIMIT or LIMIT == 'show':
            await self._show_limits(TORRENT_FILTER, directions)
        else:
            # Do we adjust current limits or set absolute limits?
            limit = LIMIT.strip()
            if limit[:2] == '+=' or limit[:2] == '-=':
                adjust = True
                limit = limit[0] + limit[2:]  # Remove '=' so it can be parsed as a number
            else:
                adjust = False

            await self._set_limits(TORRENT_FILTER, directions, limit,
                                   adjust=adjust, quiet=quiet)

    async def _show_global_limits(self, directions):
        for d in directions:
            get_method = getattr(self.srvapi.settings, 'get_limit_rate_' + d)
            limit = await get_method()
            self._output('Global %sload rate limit: %s' % (d, limit))

    async def _show_individual_limits(self, TORRENT_FILTER, directions):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        request = self.srvapi.torrent.torrents(
            tfilter, keys=('name', 'limit-rate-up', 'limit-rate-down'))
        response = await self.make_request(request, polling_frenzy=True, quiet=True)
        if response.success:
            for t in response.torrents:
                for d in directions:
                    self._output('%s %sload rate limit: %s' %
                                 (t['name'], d, t['limit-rate-%s' % d]))

    async def _set_global_limits(self, directions, limit, quiet=False, adjust=False):
        for d in directions:
            log.debug('Setting global %s rate limit: %r', d, limit)
            set_method = getattr(self.srvapi.settings,
                                 ('adjust' if adjust else 'set') + '_limit_rate_' + d)
            get_method = getattr(self.srvapi.settings, 'get_limit_rate_' + d)
            try:
                try:
                    await set_method(limit)
                except ValueError as e:
                    raise CmdError('%s: %r' % (e, limit))
                if not quiet:
                    limit = await get_method()
                    self.info('Global %sload rate limit: %s' % (d, limit))
            except ClientError as e:
                raise CmdError(e)

    async def _set_individual_limits(self, TORRENT_FILTER, directions, limit, quiet=False, adjust=False):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)

        log.debug('Setting %sload rate limit for %s torrents: %r',
                  '+'.join(directions), tfilter, limit)

        success = True
        for d in directions:
            method = getattr(self.srvapi.torrent,
                             ('adjust' if adjust else 'set') + '_limit_rate_' + d)
            response = await self.make_request(method(tfilter, limit),
                                               polling_frenzy=True, quiet=quiet)
            success = success and response.success
        if not success:
            raise CmdError()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

# Remove python from process name when running inside tmux
import os
if 'TMUX' in os.environ:
    try:
        from setproctitle import setproctitle
    except ImportError:
        pass
    else:
        from . import __appname__
        setproctitle(__appname__)


import sys
import asyncio
aioloop = asyncio.get_event_loop()


from . import cliopts
cliargs, clicmds = cliopts.parse()


from . import logging
logging.setup(debugmods=cliargs['debug'], filepath=cliargs['debug_file'])
logging.redirect_level('INFO', sys.stdout)
log = logging.make_logger(__name__)


from . import settings
localcfg = settings.Settings()
settings.init_defaults(localcfg)


from .helpmgr import HelpManager
helpmgr = HelpManager()
helpmgr.localcfg = localcfg


from .client import API
srvapi = API(host=localcfg['connect.host'],
             port=localcfg['connect.port'],
             path=localcfg['connect.path'],
             user=localcfg['connect.user'],
             password=localcfg['connect.password'],
             tls=localcfg['connect.tls'],
             interval=localcfg['tui.poll'],
             loop=aioloop)
remotecfg = srvapi.settings
helpmgr.remotecfg = remotecfg


from .client import geoip
if geoip.available:
    geoip.cachedir = localcfg['geoip.dir']
else:
    localcfg['geoip'] = False
geoip.enabled = localcfg['geoip']


from .commands import CommandManager
cmdmgr = CommandManager(loop=aioloop,
                        info_handler=lambda msg: log.info(msg),
                        error_handler=lambda msg: log.error(msg))
cmdmgr.resources.update(aioloop=aioloop,
                        srvapi=srvapi,
                        cfg=localcfg,
                        srvcfg=srvapi.settings,
                        helpmgr=helpmgr)
cmdmgr.load_cmds_from_module(
    'stig.commands.cli', 'stig.commands.tui',
)
helpmgr.cmdmgr = cmdmgr


def _pre_run_hook(cmdline):
    # Change command before it is executed

    # If there is '-h' or '--help' in the arguments, replace it with 'help
    # <cmd>'.  This is dirty but easier than forcing argparse to ignore all
    # other arguments without calling sys.exit().
    if '-h' in cmdline or '--help' in cmdline:
        cmdcls = cmdmgr.get_cmdcls(cmdline[0], interface='ANY')
        if cmdcls is not None:
            if cmdcls.name != 'tab':
                return ['help', cmdcls.name]
            else:
                # 'tab ls -h' is a little trickier because both 'tab' and 'ls'
                # can have arbitrary additional arguments which we must remove.
                #
                # Find first argument to 'tab' that is also a valid command
                # name.  Preserve all arguments before that.
                tab_args = []
                for arg in cmdline[1:]:
                    if cmdmgr.get_cmdcls(arg, interface='ANY') is not None:
                        return ['tab'] + tab_args + ['help', arg]
                    else:
                        tab_args.append(arg)
                return ['help', 'tab']
    return cmdline
cmdmgr.pre_run_hook = _pre_run_hook



def run():
    from .commands.guess_ui import (guess_ui, UIGuessError)
    from .commands import CmdError
    from . import hooks

    # Read commands from rc file
    rclines = ()
    if not cliargs['norcfile']:
        from .settings import rcfile
        from .settings.defaults import DEFAULT_RCFILE
        try:
            rclines = rcfile.read(cliargs['rcfile'] or DEFAULT_RCFILE)
        except rcfile.RcFileError as e:
            log.error('Loading rc file failed: {}'.format(e))
            sys.exit(1)

    # Decide if we run as a TUI or CLI
    if cliargs['tui']:
        cmdmgr.active_interface = 'tui'
    elif cliargs['notui']:
        cmdmgr.active_interface = 'cli'
    else:
        try:
            cmdmgr.active_interface = guess_ui(clicmds, cmdmgr)
        except UIGuessError as e:
            log.error('Unable to guess user interface')
            log.error('Provide one of these options: --tui/-t or --no-tui/-T')
            sys.exit(1)
        except CmdError as e:
            log.error(e)
            sys.exit(1)

    def run_commands():
        for cmdline in rclines:
            success = cmdmgr.run_sync(cmdline)
            # Ignored commands return None, which we consider a success here
            # because TUI commands like 'tab' in the rc file should have no
            # effect at all when in CLI mode.
            if success is False:
                return False

        # Exit if CLI commands fail
        if clicmds:
            success = cmdmgr.run_sync(clicmds)
            if not success:
                return False

        return True

    exit_code = 0

    # Run commands either in CLI or TUI mode
    if cmdmgr.active_interface == 'cli':
        # Exit when pipe is closed (e.g. `stig help | head -1`)
        import signal
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)

        # Load geoip database
        if localcfg['geoip']:
            try:
                aioloop.run_until_complete(geoip.load(loop=aioloop))
            except geoip.GeoIPError as e:
                log.error(e)
                exit_code = 1

        try:
            if not run_commands():
                exit_code = 1
        except KeyboardInterrupt:
            log.debug('Caught SIGINT')

    elif cmdmgr.active_interface == 'tui':
        from .tui import main as tui
        cmdmgr.resources.update(tui=tui)
        if not tui.run(run_commands):
            exit_code = 1

    # Terminate any remaining tasks
    tasks = tuple(task for task in asyncio.Task.all_tasks() if not task.done())
    if tasks:
        log.debug('Not all tasks have been properly canceled.')
        for task in tasks:
            log.debug('Terminating leftover task: %r', task)
            task.cancel()
            try:
                aioloop.run_until_complete(asyncio.wait_for(task, timeout=None))
                task.result()
            except (asyncio.CancelledError, asyncio.TimeoutError):
                pass

    _cancel_unfinished_tasks()
    aioloop.run_until_complete(srvapi.rpc.disconnect('Quit'))
    # # Closing the event loop raises "RuntimeError: Event loop is closed" (not
    # # always) when a `run_in_executor` command (i.e. a thread) is cancelled.
    # # https://github.com/python/asyncio/issues/258
    # aioloop.close()
    sys.exit(exit_code)


def _cancel_unfinished_tasks():
    pending = (task for task in asyncio.Task.all_tasks() if not task.done())
    for task in pending:
        log.debug('Cancelling pending task: %r', task)
        task.cancel()
        try:
            log.debug('Finishing pending task: %r', task)
            aioloop.run_until_complete(task)
        except asyncio.CancelledError:
            pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..base import peer as base
from . import _mixin as mixin


class ListPeersCmd(base.ListPeersCmdbase,
                   mixin.make_request,
                   mixin.select_torrents,
                   mixin.create_list_widget):
    provides = {'tui'}

    def make_peer_list(self, tfilter, pfilter, sort, columns):
        from ...tui.views.peer_list import PeerListWidget
        self.create_list_widget(PeerListWidget, theme_name='peerlist',
                                tfilter=tfilter, pfilter=pfilter,
                                sort=sort, columns=columns,
                                markable_items=False)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

from functools import partial
import os

from ..main import (localcfg, srvapi, aioloop)
from . import main as tui

from .views.torrent_list import TorrentListWidget
from .views.file_list import FileListWidget
from .views.peer_list import PeerListWidget

from .views.torrent import TUICOLUMNS as TORRENT_COLUMNS
from .views.file import TUICOLUMNS as FILE_COLUMNS


def _reconnect(settings, name, value):
    # See also ..hooks
    log.debug('Reconnecting because %s changed to %r', name, value)
    srvapi.poll()
localcfg.on_change(_reconnect, name='connect.host')
localcfg.on_change(_reconnect, name='connect.port')
localcfg.on_change(_reconnect, name='connect.path')
localcfg.on_change(_reconnect, name='connect.user')
localcfg.on_change(_reconnect, name='connect.password')
localcfg.on_change(_reconnect, name='connect.tls')


def _update_pollers(rpc):
    tui.srvapi.poll()
srvapi.rpc.on('connected', _update_pollers)
srvapi.rpc.on('disconnected', _update_pollers)


def _refresh_lists(settings, name, value):
    for widget in tui.tabs:
        if isinstance(widget, (TorrentListWidget, FileListWidget, PeerListWidget)):
            widget.clear()
            widget.refresh()
localcfg.on_change(_refresh_lists, name='unit.bandwidth')
localcfg.on_change(_refresh_lists, name='unit.size')
localcfg.on_change(_refresh_lists, name='unitprefix.bandwidth')
localcfg.on_change(_refresh_lists, name='unitprefix.size')
localcfg.on_change(_refresh_lists, name='reverse-dns')
localcfg.on_change(_refresh_lists, name='geoip')


def _set_poll_interval(settings, name, value):
    tui.srvapi.interval = value
localcfg.on_change(_set_poll_interval, name='tui.poll')


def _set_cli_history_dir(settings, name, value):
    tui.cli.original_widget.history_file = os.path.join(value, 'commands')
localcfg.on_change(_set_cli_history_dir, name='tui.cli.history-dir')

def _set_cli_history_size(settings, name, value):
    tui.cli.original_widget.history_size = value
localcfg.on_change(_set_cli_history_size, name='tui.cli.history-size')


def _set_autohide_delay(settings, name, value):
    tui.logwidget.autohide_delay = value
localcfg.on_change(_set_autohide_delay, name='tui.log.autohide')


def _set_log_height(settings, name, value):
    tui.logwidget.height = int(value)
localcfg.on_change(_set_log_height, name='tui.log.height')


def _set_theme(settings, name, value):
    try:
        tui.load_theme(value)
    except tui.theme.ThemeError as e:
        raise ValueError(e)
localcfg.on_change(_set_theme, name='tui.theme')


def _set_tui_marked_char(methodname, settings, name, value):
    getattr(TORRENT_COLUMNS['marked'], methodname)(value)
    getattr(FILE_COLUMNS['marked'], methodname)(value)
    for widget in tui.tabs:
        if isinstance(widget, (TorrentListWidget, FileListWidget)):
            widget.refresh_marks()
localcfg.on_change(partial(_set_tui_marked_char, 'set_marked_char'), name='tui.marked.on', autoremove=False)
localcfg.on_change(partial(_set_tui_marked_char, 'set_unmarked_char'), name='tui.marked.off', autoremove=False)
_set_tui_marked_char('set_marked_char', localcfg, name='tui.marked.on', value=localcfg['tui.marked.on'])
_set_tui_marked_char('set_unmarked_char', localcfg, name='tui.marked.off', value=localcfg['tui.marked.off'])


def _update_quickhelp(keymap):
    tui.topbar.help.update()
tui.keymap.on_bind_unbind(_update_quickhelp)


def _set_geoip(settings, name, value):
    if value:
        tui.load_geoip_db()
localcfg.on_change(_set_geoip, name='geoip')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes for CLI commands"""

from ...logging import make_logger
log = make_logger(__name__)

from .. import ExpectedResource
from .. import utils
from ._common import clear_line


class make_request():
    async def make_request(self, request_coro, polling_frenzy=False, quiet=False):
        """Awaits request coroutine and logs messages; returns response"""
        response = await request_coro
        utils.log_msgs(self, response, quiet)
        return response


class ask_yes_no():
    aioloop = ExpectedResource

    ANSWERS = {'y': True, 'n': False,
               'Y': True, 'N': False,
               '\x03': False,  # ctrl-c
               '\x07': False,  # ctrl-g
               '\x1b': False}  # escape

    async def ask_yes_no(self, question, yes=None, no=None, after=None):
        """
        Ask user a yes/no question and return True/False

        The `yes` and `no` arguments are callbacks (or None) that are called
        depending on the user's answer. `after` is called after the user
        answered and after `yes` or `no` has been called.

        Callbacks may be normal functions, coroutine functions or
        coroutines. They don't get any arguments and their return value is
        ignored.
        """
        answer = await self._get_answer(question)
        if answer:
            await self._run_func_or_coro(yes)
        else:
            await self._run_func_or_coro(no)
        await self._run_func_or_coro(after)
        return answer

    async def _run_func_or_coro(self, func_or_coro):
        import asyncio
        if asyncio.iscoroutinefunction(func_or_coro):
            await func_or_coro()
        elif asyncio.iscoroutine(func_or_coro):
            await func_or_coro
        elif func_or_coro is not None:
            func_or_coro()

    async def _get_answer(self, question):
        import sys, tty, termios

        if not sys.stdout.isatty():
            # We can't ask the user - default to None (which evaluates to False)
            return None

        async def aiogetch(loop):
            # Disable printing of typed characters
            old_settings = termios.tcgetattr(sys.stdin.fileno())
            tty.setraw(sys.stdin.fileno())

            # Read exactly one character
            key = await loop.run_in_executor(None, sys.stdin.read, 1)

            # Restore terminal settings
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, old_settings)

            return key

        # Get answer from user
        answer = None
        while answer is None:
            print(question, end=' [y|n] ', flush=True)
            key = await aiogetch(self.aioloop)
            clear_line()
            answer = self.ANSWERS.get(key, None)
        return answer


class select_torrents():
    def select_torrents(self, FILTER, allow_no_filter=True, discover_torrent=None, prefer_focused=None):
        """
        Get TorrentFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False, None is returned if allow_no_filter
        evaluates to True, otherwise a ValueError is raised.

        `discover_torrent` and `prefer_focused` are ignored and only used in the
        TUI version of this method.
        """
        if FILTER:
            from ...client import TorrentFilter
            return TorrentFilter(FILTER)
        else:
            if allow_no_filter:
                return None
            else:
                raise ValueError('No torrent specified')


class select_files():
    def get_focused_path_in_torrent(self):
        """Return relative path in torrent of focused file or directory"""
        pass

    def select_files(self, FILTER, allow_no_filter=True, discover_file=None):
        """
        Get TorrentFileFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentFileFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False, None is returned if allow_no_filter
        evaluates to True, otherwise a ValueError is raised.

        `discover_file` is ignored and only used in the TUI version of this
        method (see ..tui.mixin.select_file).
        """
        if FILTER:
            from ...client import TorrentFileFilter
            return TorrentFileFilter(FILTER)
        else:
            if allow_no_filter:
                return None
            else:
                raise ValueError('No torrent specified')


class only_supported_columns():
    def only_supported_columns(self, columns, specs):
        """Remove columns from list `columns` that don't support the CLI"""
        return [col for col in columns
                if 'cli' in specs[col].interfaces]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes that are common between TUI and CLI commands"""


class get_single_torrent():
    async def get_single_torrent(self, tfilter, keys=()):
        """
        Get a single torrent that matches TorrentFilter `tfilter`

        If there are multple matches, they are sorted by name and the first
        match is returned.

        Return None if no match is found.
        """
        request = self.srvapi.torrent.torrents(tfilter, keys=tuple(keys) + ('name',))
        response = await self.make_request(request, polling_frenzy=False, quiet=True)
        if response.success:
            from ...client import TorrentSorter
            torrents = TorrentSorter(('name',)).apply(response.torrents)
            return torrents[0]

class get_torrent_sorter():
    def get_torrent_sorter(self, args):
        """
        Return TorrentSorter instance or None

        If `args` evaluates to True, it is passed to TorrentSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TorrentSorter
            return TorrentSorter(self.cfg.validate('sort.torrents', args))

class get_torrent_columns():
    def get_torrent_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid torrent list column name

        Raise ValueError or return a new list of `columns`.
        """
        return self.cfg.validate('columns.torrents', columns)



class get_file_columns():
    def get_file_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid file list column name

        Raise ValueError or return a new list of `columns`.
        """
        return self.cfg.validate('columns.files', columns)



class get_peer_sorter():
    def get_peer_sorter(self, args):
        """
        Return TorrentPeerSorter instance or None

        If `args` evaluates to True, it is passed to TorrentPeerSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TorrentPeerSorter
            return TorrentPeerSorter(self.cfg.validate('sort.peers', args))

class get_peer_filter():
    def get_peer_filter(self, FILTER):
        """
        Return TorrentPeerFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentPeerFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False, None is returned.
        """
        if FILTER:
            from ...client import TorrentPeerFilter
            return TorrentPeerFilter(FILTER)

class get_peer_columns():
    def get_peer_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid peer list column name

        Raise ValueError or return a new list of `columns`.
        """
        cols = self.cfg.validate('columns.peers', columns)
        if 'country' in cols:
            from ...main import localcfg
            if not localcfg['geoip']:
                cols = cols.copy(*(col for col in cols if col != 'country'))
        return cols



class get_tracker_sorter():
    def get_tracker_sorter(self, args):
        """
        Return TorrentTrackerSorter instance or None

        If `args` evaluates to True, it is passed to TorrentTrackerSorter and
        the result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TorrentTrackerSorter
            return TorrentTrackerSorter(self.cfg.validate('sort.trackers', args))

class get_tracker_filter():
    def get_tracker_filter(self, FILTER):
        """
        Return TorrentTrackerFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentTrackerFilter and
        the resulting object is returned.

        If `FILTER` evaluates to False, None is returned.
        """
        if FILTER:
            from ...client import TorrentTrackerFilter
            return TorrentTrackerFilter(FILTER)

class get_tracker_columns():
    def get_tracker_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid tracker list column name

        Raise ValueError or return a new list of `columns`.
        """
        return self.cfg.validate('columns.trackers', columns)



class get_setting_sorter():
    def get_setting_sorter(self, args):
        """
        Return SettingSorter instance or None

        If `args` evaluates to True, it is passed to SettingSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import SettingSorter
            return SettingSorter(self.cfg.validate('sort.settings', args))

class get_setting_columns():
    def get_setting_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid setting list column name

        Raise ValueError or return a new list of `columns`.
        """
        return self.cfg.validate('columns.settings', columns)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from .. import (InitCommand, CmdError, ExpectedResource)
from . import _mixin as mixin
from ._common import (make_X_FILTER_spec, make_COLUMNS_doc,
                      make_SORT_ORDERS_doc, make_SCRIPTING_doc)

import asyncio
from collections import abc


class ListPeersCmdbase(mixin.get_peer_sorter, mixin.get_peer_columns,
                       mixin.get_peer_filter, metaclass=InitCommand):
    name = 'peerlist'
    aliases = ('pls', 'lsp')
    provides = set()
    category = 'peer'
    description = 'List connected peers of torrent(s)'
    usage = ('peerlist [<OPTIONS>]',
             'peerlist [<OPTIONS>] <TORRENT FILTER>',
             'peerlist [<OPTIONS>] <TORRENT FILTER> <PEER FILTER>')
    examples = ('peerlist',
                'peerlist downloading',
                'peerlist some_torrent host=127.0.0.1')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),
        make_X_FILTER_spec('PEER', nargs='?'),

        { 'names': ('--sort', '-s'),
          'default_description': "current value of 'sort.peers' setting",
          'description': ('Comma-separated list of sort orders '
                          "(see SORT ORDERS section)") },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'columns.peers' setting",
          'description': ('Comma-separated list of column names '
                          "(see COLUMNS section)") },
    )

    from ...views.peer import COLUMNS
    from ...client.sorters.peer import TorrentPeerSorter
    more_sections = {
        'COLUMNS': make_COLUMNS_doc(COLUMNS, '--columns', 'columns.peers', append=(
            '',
            'The "torrent" column is added automatically if multiple '
            'torrents could be listed potentially.')),
        'SORT ORDERS': make_SORT_ORDERS_doc(TorrentPeerSorter, '--sort', 'sort.peers'),
        'SCRIPTING': make_SCRIPTING_doc(name),
    }

    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, PEER_FILTER, sort, columns):
        columns = self.cfg['columns.peers'] if columns is None else columns
        sort = self.cfg['sort.peers'] if sort is None else sort
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=True,
                                           discover_torrent=True)
            pfilter = self.get_peer_filter(PEER_FILTER)
            sort    = self.get_peer_sorter(sort)
            columns = self.get_peer_columns(columns)
        except ValueError as e:
            raise CmdError(e)

        # Unless we're listing peers of exactly one torrent, specified by its
        # ID, automatically add the 'torrent' column.
        if 'torrent' not in columns and \
           (not isinstance(tfilter, abc.Sequence) or len(tfilter) != 1):
            columns += ('torrent',)

        log.debug('Listing %s peers of %s torrents', pfilter, tfilter)

        if asyncio.iscoroutinefunction(self.make_peer_list):
            await self.make_peer_list(tfilter, pfilter, sort, columns)
        else:
            self.make_peer_list(tfilter, pfilter, sort, columns)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from .__version__ import __version__
__appname__ = __name__.split('.')[0]
__url__ = 'https://github.com/rndusr/stig'

def run():
    try:
        from . import main
        if main.cliargs['profile_file'] is not None:
            main.logging.start_profiling(main.run,
                                         filepath=main.cliargs['profile_file'],
                                         statistical=False)
        else:
            main.run()
    except KeyboardInterrupt:
        pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import urwid

from .setting import TUICOLUMNS
from . import (ItemWidgetBase, ListWidgetBase)
from ...main import (localcfg, remotecfg, srvapi, aioloop)
from ...utils.usertypes import (Bool, Option)
from ...client import SettingFilter


def _change_setting(name, new_value, on_success=None):
    remote_name = name[4:]  # Remove 'srv.'

    if name in localcfg:
        try:
            localcfg[name] = new_value
        except ValueError as e:
            log.error('Cannot set %s = %r: %s', name, new_value, e)
        else:
            if on_success is not None:
                on_success()

    elif remote_name in remotecfg:
        async def setter():
            try:
                await remotecfg.set(remote_name, new_value)
            except (ValueError, srvapi.ClientError) as e:
                log.error('Cannot set %s = %r: %s', name, new_value, e)
            else:
                if on_success is not None:
                    on_success()
        aioloop.create_task(setter())

    else:
        raise RuntimeError('Not a setting name: %r' % name)


class SettingItemWidget(ItemWidgetBase):
    palette_unfocused = 'settinglist'
    palette_focused   = 'settinglist.focused'
    columns_focus_map = {}
    for col in TUICOLUMNS.values():
        columns_focus_map.update(col.style.focus_map)

    @property
    def id(self):
        return self.data['id']

    def selectable(self):
        return True

    @property
    def name(self):
        if self._cells.exists('name'):
            return self._cells.name.text.text

    @property
    def value_widget(self):
        if self._cells.exists('value'):
            return self._cells.value.base_widget

    @property
    def current_value(self):
        if self._cells.exists('value'):
            value_widget = self.value_widget
            if isinstance(value_widget, urwid.Edit):
                return value_widget.edit_text
            elif hasattr(value_widget, 'get_tui_value'):
                return value_widget.get_tui_value()

    @property
    def edit_mode(self):
        if self._cells.exists('value'):
            return isinstance(self._cells.value.base_widget, urwid.Edit)
        return False

    def keypress(self, size, key):
        current_value = self.current_value
        if current_value is None:
            return key
        elif isinstance(current_value, Bool):
            return self._keypress_bool(size, key)
        elif isinstance(current_value, Option):
            return self._keypress_option(size, key)
        else:
            return self._keypress_string(size, key)

    def _keypress_bool(self, size, key):
        cmd = self._command_map[key]
        if cmd is urwid.ACTIVATE:
            new_value = not self.current_value
            _change_setting(self.name, new_value)
        else:
            return key

    def _keypress_option(self, size, key):
        cmd = self._command_map[key]
        if cmd is urwid.ACTIVATE:
            current_value = self.current_value
            options = current_value.options
            index = options.index(current_value)
            if index < len(options)-1:
                index += 1
            else:
                index = 0
            new_value = options[index]
            _change_setting(self.name, new_value)
        else:
            return key

    def _keypress_string(self, size, key):
        cells = self._cells
        cmd = self._command_map[key]
        current_value = self.current_value
        value_widget = self.value_widget

        def edit():
            attrmap = self._cells.value.attrmap
            edit_widget = urwid.AttrMap(urwid.Edit(edit_text=str(current_value)),
                                        attr_map=attrmap.attr_map,
                                        focus_map=attrmap.focus_map)
            cells.replace('value', edit_widget)
            self._value_widget_temp = value_widget

        def unedit():
            cells.replace('value', self._value_widget_temp)
            delattr(self, '_value_widget_temp')

        if cmd is urwid.ACTIVATE:
            if not self.edit_mode:
                edit()
            else:
                new_value = value_widget.edit_text
                _change_setting(self.name, new_value, on_success=unedit)
        elif cmd is urwid.CANCEL:
            if self.edit_mode:
                unedit()
        elif self.edit_mode:
            key = super().keypress(size, key)
            cmd = self._command_map[key]
            # Don't allow user to focus next/previous setting when editing
            if cmd not in (urwid.CURSOR_DOWN, urwid.CURSOR_UP):
                return key
        else:
            return key


class SettingListWidget(ListWidgetBase):
    tuicolumns      = TUICOLUMNS
    ListItemClass   = SettingItemWidget
    keymap_context  = 'setting'
    palette_name    = 'settinglist'
    focusable_items = True

    def __init__(self, srvapi, keymap, sort=None, columns=None, title='Settings'):
        super().__init__(srvapi, keymap, columns=columns, sort=sort, title=title)
        self._sort = sort
        self._secondary_filter = None
        localcfg.on_change(self._handle_update)
        remotecfg.on_update(self._handle_update)
        self.refresh()

    def _handle_update(self, *_, **__):
        self._data_dict = {
            **{k: {'id': k,
                   'value': v,
                   'default': localcfg.default(k),
                   'description': localcfg.description(k)}
               for k,v in localcfg.items()},
            **{'srv.'+k: {'id': 'srv.'+k,
                          'value': v,
                          'default': '',
                          'description': remotecfg.description(k)}
               for k,v in remotecfg.items()},
        }
        self._invalidate()

    def refresh(self):
        remotecfg.poll()

    @property
    def sort(self):
        return self._sort

    @sort.setter
    def sort(self, sort):
        ListWidgetBase.sort.fset(self, sort)
        self.refresh()

    @property
    def secondary_filter(self):
        return self._secondary_filter

    @secondary_filter.setter
    def secondary_filter(self, setting_filter):
        if setting_filter is None:
            self._secondary_filter = None
        else:
            self._secondary_filter = SettingFilter(setting_filter)
        self._invalidate()

    def _limit_items(self, setting_widgets):
        sfilter = self._secondary_filter
        if sfilter is not None:
            for sw in setting_widgets:
                if not sfilter.match(sw.data):
                    yield sw




# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""General hooks that are always needed regardless of the interface"""

from .logging import make_logger
log = make_logger(__name__)

from .main import (localcfg, srvapi, geoip)
from .utils import convert
from .views.torrent import COLUMNS as TORRENT_COLUMNS
from .views.file import COLUMNS as FILE_COLUMNS
from .views.peer import COLUMNS as PEER_COLUMNS


def _make_connection_callback(attr):
    def on_set(settings, name, value):
        log.debug('Setting rpc.%s=%r', attr, value)
        setattr(srvapi.rpc, attr, value)
    return on_set
localcfg.on_change(_make_connection_callback('host'),     name='connect.host',     autoremove=False)
localcfg.on_change(_make_connection_callback('port'),     name='connect.port',     autoremove=False)
localcfg.on_change(_make_connection_callback('path'),     name='connect.path',     autoremove=False)
localcfg.on_change(_make_connection_callback('user'),     name='connect.user',     autoremove=False)
localcfg.on_change(_make_connection_callback('password'), name='connect.password', autoremove=False)
localcfg.on_change(_make_connection_callback('tls'),      name='connect.tls',      autoremove=False)
localcfg.on_change(_make_connection_callback('timeout'),  name='connect.timeout',  autoremove=False)


_BANDWIDTH_COLUMNS = (TORRENT_COLUMNS['rate-up'], TORRENT_COLUMNS['rate-down'],
                      TORRENT_COLUMNS['limit-rate-up'], TORRENT_COLUMNS['limit-rate-down'],
                      PEER_COLUMNS['rate-up'], PEER_COLUMNS['rate-down'], PEER_COLUMNS['rate-est'])
def _set_bandwidth_unit(settings, name, value):
    convert.bandwidth.unit = value
    unit_short = convert.bandwidth.unit
    for column in _BANDWIDTH_COLUMNS:
        column.set_unit(unit_short)
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_bandwidth_unit, name='unit.bandwidth')
_set_bandwidth_unit(localcfg, name='unit.bandwidth', value=localcfg['unit.bandwidth'])  # Init columns' units

def _set_bandwidth_prefix(settings, name, value):
    convert.bandwidth.prefix = value
    for column in _BANDWIDTH_COLUMNS:
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_bandwidth_prefix, name='unitprefix.bandwidth')


_SIZE_COLUMNS = (TORRENT_COLUMNS['size'], TORRENT_COLUMNS['downloaded'],
                 TORRENT_COLUMNS['uploaded'], TORRENT_COLUMNS['available'],
                 FILE_COLUMNS['size'], FILE_COLUMNS['downloaded'])
def _set_size_unit(settings, name, value):
    convert.size.unit = value
    unit_short = convert.size.unit
    for column in _SIZE_COLUMNS:
        column.set_unit(unit_short)
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_size_unit, name='unit.size')
_set_size_unit(localcfg, name='unit.size', value=localcfg['unit.size'])  # Init columns' units

def _set_size_prefix(settings, name, value):
    convert.size.prefix = value
    for column in _SIZE_COLUMNS:
        column.clearcache()
    srvapi.torrent.clearcache()
    srvapi.poll()
localcfg.on_change(_set_size_prefix, name='unitprefix.size')


def _set_geoip(settings, name, value):
    if value and not geoip.available:
        log.error('Missing geoip dependency: maxminddb')
        localcfg['geoip'] = value = False
    geoip.enabled = value
localcfg.on_change(_set_geoip, name='geoip')

def _set_geoip_dir(settings, name, value):
    geoip.cachedir = value
localcfg.on_change(_set_geoip_dir, name='geoip.dir')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import re
import blinker


class Key(str):
    """Key or key combination as string"""

    # Convert some urwid key names and some other stuff
    _INIT = (
        (re.compile(r'^<(.+)>$'),                            r'\1'),
        (re.compile(r'^(.*) $'),                             r'\1space'),
        (re.compile(r'^meta', flags=re.I),                   r'alt'),
        (re.compile(r'\besc$', flags=re.I),                  r'escape'),
        (re.compile(r'\bpos1$', flags=re.I),                 r'home'),
        (re.compile(r'\bdel$', flags=re.I),                  r'delete'),
        (re.compile(r'\bins$', flags=re.I),                  r'insert'),
        (re.compile(r'(\b|\W*)(?:return|\n)$', flags=re.I),  r'\1enter'),
        (re.compile(r'\bpage up$', flags=re.I),              r'pgup'),
        (re.compile(r'\bpage down$', flags=re.I),            r'pgdn'),
        (re.compile(r'\bpage dn$', flags=re.I),              r'pgdn'),
        (re.compile(r' '),                                   r'-'),
        # The first part in key combos must always be the same, but the part
        # after must be preserved. <alt-l> is not the same as <alt-L>.
        (re.compile(r'^(\w+)-(\S+)$'),
         lambda m: m.group(1).lower()+'-'+m.group(2)),
    )

    _MODS = ('shift', 'alt', 'ctrl')
    _FKEYS = ('F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12',
              'F13', 'F14', 'F15', 'F16', 'F17', 'F18', 'F19', 'F20')
    _KEYNAMES = ('escape', 'space', 'home', 'end', 'tab', 'delete', 'backspace', 'insert',
                 'enter', 'pgup', 'pgdn', 'left', 'right', 'up', 'down') + _FKEYS
    _cache = {}

    def __new__(cls, key):
        if isinstance(key, Key):
            return key
        elif key in cls._cache:
            return cls._cache[key]
        elif not isinstance(key, str):
            raise TypeError('Not a string: %r' % key)
        else:
            orig_key = key

        # Remove brackets (<>) around key, some renaming, etc.
        for regex,repl in cls._INIT:
            key = regex.sub(repl, key)

        # Split modifiers and key
        if '-' in key and len(key) > 1:
            _parts = key.split('-')
            key = _parts.pop(-1)
            mods = set(p.lower() for p in _parts)
            for mod in mods:
                if mod not in cls._MODS:
                    raise ValueError('Invalid modifier in <%s>: <%s>' % (orig_key, mod))

            if len(key) == 0:
                raise ValueError('Missing key after modifier: <%s>' % orig_key)

            # Convert <shift-x> to <X>
            elif len(key) == 1 and 'shift' in mods:
                key = key.upper()
                mods.remove('shift')

            # 'shift-/ctrl-E' is the same as 'shift-/ctrl-e'
            elif 'alt' not in mods:
                key = key.lower()
        else:
            mods = set()

        # Validate named keys ('enter', 'delete', etc)
        if len(key) > 1:
            # F* key names are upper case, all others lower case
            key_upper = key.upper()
            key = key_upper if key_upper in cls._FKEYS else key.lower()
            if key not in cls._KEYNAMES:
                raise ValueError('Unknown key: %r' % key)

        if mods:
            keystr = '%s-%s' % ('-'.join(sorted(mods)), key)
        else:
            keystr = key

        if len(key) < 1:
            raise ValueError('No key specified')

        obj = super().__new__(cls, keystr)
        cls._cache[orig_key] = obj
        return obj

    def __eq__(self, other):
        if not isinstance(other, Key):
            try:
                other = Key(other)
            except (ValueError, TypeError):
                return False
        return super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __str__(self):
        return '<%s>' % super().__str__()

    def __repr__(self):
        return '<Key %s>' % super().__str__()


class KeyChain(tuple):
    ADVANCE = '<ADVANCE>'
    REDUCE  = '<REDUCE>'
    REJECT  = '<REJECT>'
    ABORT   = '<ABORT>'

    def __new__(cls, *keys):
        obj = super().__new__(cls, (Key(k) for k in keys))
        if len(keys) < 2:
            raise ValueError('Not enough keys to make a chain: %s' % str(obj))
        return obj

    def startswith(self, keys):
        """Whether this key chain starts with `keys` (sequence of Key instances)"""
        keys_ = tuple(keys)
        len_keys_ = len(keys_)
        return len_keys_ <= len(self) and self[:len_keys_] == keys_

    def __eq__(self, other):
        return super().__eq__(tuple(other))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __str__(self):
        return ' '.join(str(k) for k in self)

    def __repr__(self):
        return '<KeyChain %s>' % str(self)


class Keymapped():
    """Mixin class that provides a keypress method"""

    def keypress(self, size, key_orig):
        keymap   = self.__keymap
        context  = self.__context
        callback = self.__callback
        key_eval = Key(key_orig)

        def try_parent_class(key, sup=super()):
            if sup.selectable():
                # mro = sup.__self_class__.mro()
                # super_cls = mro[mro.index(sup.__thisclass__) + 1]
                # log.debug('%s: Offering %r to %s.keypress()', context, key, super_cls.__name__)
                key = sup.keypress(size, key)
                # log.debug('%s: %s.keypress() returned %r', context, super_cls.__name__, key)
            return key

        log.debug('%s: %r / %r pressed', context, key_orig, key_eval)

        # Any started keychains have the highest priority
        if keymap.any_keychain_started:
            log.debug('%s: Advancing started keychain(s)', context)
            key_ = keymap.evaluate(key_eval, context=context, callback=callback, widget=self)
            if key_ is None:
                log.debug('%s: Started keychain(s) used %r', context, key_eval)
                return None

        # Offer original key to original keypress() for built-in keys like up/down
        key_ = try_parent_class(key_orig)
        if key_ is None:
            return None

        # Offer our own key format to original keypress() in case there are any
        # Key instances in urwid.command_map
        if str(key_orig) != str(key_eval):
            key_ = try_parent_class(key_eval)
            if key_ is None:
                return None
        else:
            log.debug('%s: Not feeding %r to parent again (%r == %r)', context, key_eval, key_orig, key_eval)

        # Offer key to KeyMap
        key_ = keymap.evaluate(key_eval, context=context, callback=callback, widget=self)
        if key_ is None:
            # Key was evaluated() to some action
            return None
        elif key_ != key_eval:
            # key_eval was translated to another key (e.g. 'j' -> 'down'), so we
            # check if original keypress() likes that better
            if try_parent_class(key_) is None:
                # Original keypress() used translated key
                return None

        log.debug('%s: Returning %r', context, key_eval)
        return key_eval

    def selectable(self):
        return True


class KeyMap():
    """
    Bind keys to actions in different contexts

    To bind keys to widgets, a new class is created that accepts user input by
    adding the methods `keypress` and `selectable` to the original class:

    >>> keymap = KeyMap()
    >>> keymap.bind(key='meta g',
                    action=lambda widget: widget.set_edit_text('abcfoo'),
                    context='pwentry')
    >>> Password = keymap.wrap(urwid.Edit, context='pwentry')
    >>> Password
    urwid.Edit_Keymapped

    Now you can use Password like any regular Edit widget:

    >>> pw = Password('', 'Enter password or <Alt-g> to generate one')
    """

    NO_CONTEXT      = object()
    ALL_CONTEXTS    = object()
    DEFAULT_CONTEXT = 'default'

    def __init__(self, callback=None):
        self._default_callback = callback
        self._actions = {self.DEFAULT_CONTEXT: {}}
        self._descriptions = {}

        self._bindunbind_callbacks = blinker.Signal()
        self._keychain_callbacks = blinker.Signal()
        self._keychain_partial = []
        self._keychain_context = self.NO_CONTEXT

    def clear(self):
        """Remove all keybindings"""
        contexts = self._actions
        for context in contexts:
            log.debug('%s: Removing all keybindings', context)
            contexts[context] = {}
        self._bindunbind_callbacks.send(self)

    def _unbind_from_urwid_command_map(self, key):
        """Remove key from urwid's internal command map if possible"""
        import urwid
        if isinstance(key, KeyChain):
            key = key[0]
        try:
            del urwid.command_map[key]
        except KeyError as e:
            return False
        else:
            return True

    def bind(self, key, action, context=DEFAULT_CONTEXT, description=None):
        """
        Bind `key` to `action` in `context`

        key: a key string that is converted to Key or KeyChain with `mkkey`
        context: a descriptive string (or any other hashable object)
        action: What to do when `key` is pressed in `context`; may be:
                - a callable that accepts the widget that received `key`,
                - a Key object (so you can bind <j> to <down> for example),
                - any other object that is passed to the `callback` specified
                  in calls to `wrap` or when the KeyMap was instantiated.
        description: a descriptive string
        """
        key = self.mkkey(key)
        if isinstance(action, Key) and key == action:
            raise ValueError('Circular key mapping: %s -> %s' % (key, action))

        if context not in self._actions:
            self._actions[context] = {}
        self._actions[context][key] = action
        self._descriptions[(context, key)] = description or ''
        log.debug('%s: Mapped %r -> %r', context, key, action)
        self._bindunbind_callbacks.send(self)

    def unbind(self, key, context=DEFAULT_CONTEXT):
        """
        Unbind `key` in `context`

        Key chains starting with `key` are also removed.
        """
        key = self.mkkey(key)
        if context not in self._actions:
            raise ValueError('Unknown context: {!r}'.format(context))
        elif key in self._actions[context]:
            del self._actions[context][key]
            log.debug('%s: Unmapped %r', context, key)
        else:
            if context == self.DEFAULT_CONTEXT:
                key_removed = self._unbind_from_urwid_command_map(key)
            else:
                key_removed = False

            for k in tuple(self._actions[context]):
                if isinstance(k, KeyChain) and k.startswith(key):
                    del self._actions[context][k]
                    log.debug('%s: Unmapped %r', context, k)
                    key_removed = True
            if not key_removed:
                raise ValueError('Key not mapped in context %r: %s' % (context, key))
        self._bindunbind_callbacks.send(self)

    def get_description(self, key, context=DEFAULT_CONTEXT):
        return self._descriptions.get((context, key), '')

    def evaluate(self, key, context=DEFAULT_CONTEXT, callback=None, widget=None):
        """
        Run action that is mapped to `key` in `context`

        key: the pressed key
        context: the context `key` was pressed in
        callback: None or any callable
        widget: the widget that received `key`
        """
        if context not in self._actions:
            raise ValueError('Unknown context: {}'.format(context))

        key = Key(key)
        action = None

        # Unless we've are already started a key chain, try to find a single-key mapping
        if not self.any_keychain_started:
            log.debug('%s: Evaluating %r as single key for widget %r', context, key, widget)
            action = self._get_single_key_action(key, context)

        # Try to advance keychains only if no keychain was started previously or
        # if that previously started keychain was in the same context as we're
        # in now.
        if action is None and (self._keychain_context == self.NO_CONTEXT or
                               self._keychain_context == context):
            log.debug('%s: Evaluating %r as chain key for widget %r', context, key, widget)
            action = self._get_keychain_action(key, context)
            if action is KeyChain.ADVANCE:
                log.debug('%s:   %r was used to advance a keychain; locking context', context, key)
                self._keychain_context = context
                self._keychain_partial.append(key)
                self._run_keychain_callbacks()
                return None
            elif action is KeyChain.REDUCE:
                log.debug('%s:   %r was used to reduce a keychain', context, key)
                self._keychain_partial.pop()
                if not self.any_keychain_started:
                    self._reset_keychains()
                else:
                    self._run_keychain_callbacks()
                return None
            elif action is KeyChain.ABORT:
                log.debug('%s:   %r was used to abort a keychain', context, key)
                self._reset_keychains()
                return None
            elif action is KeyChain.REJECT:
                log.debug('%s:   %r was rejected by all keychains', context, key)
                action = None
            elif isinstance(action, Key):
                log.debug('%s:   %r was resolved to a single key: %r', context, key, action)
                self._reset_keychains(context)
            else:
                log.debug('%s:   %r was used to complete a keychain', context, key)
                self._reset_keychains(context)

        # Handle the action we found
        if action is None:
            log.debug('%s:   %r is not mapped', context, key)
            return key
        elif isinstance(action, Key):
            return self.evaluate(action, context, callback, widget)
        elif isinstance(action, KeyChain):
            raise RuntimeError('Mapping to keychains is not supported.')
        elif callable(action):
            # action itself is the callback
            log.debug('%s:   Calling action directly: %r', context, action)
            action(widget)
        elif callback is not None:
            # Individual callback for this widget
            log.debug('%s:  Calling widget class callback %r(%r, %r)', context, callback, action, widget)
            callback(action, widget)
        elif self._default_callback is not None:
            # General callback for all widgets
            log.debug('%s:   Calling default callback %r(%r, %r)', context, self._default_callback, action, widget)
            self._default_callback(action, widget)
        else:
            raise RuntimeError('No callback given - unable to handle {!r}'.format(action))

    def _get_single_key_action(self, key, context=DEFAULT_CONTEXT):
        actions = self._actions[context]
        if key in actions:
            return actions[key]

        if context is not self.DEFAULT_CONTEXT:
            actions = self._actions[self.DEFAULT_CONTEXT]
            if key in actions:
                return actions[key]

    def _get_keychain_action(self, key, context):
        partial = self._keychain_partial
        candidate = tuple(partial) + (key,)
        log.debug('%s: Getting keychain action for %s + %s:',
                  context, '+'.join(partial) or "<>", key)
        action = KeyChain.ABORT if partial else KeyChain.REJECT
        for kc,act in self._started_keychains(context):
            if kc == candidate:
                log.debug('%s:   Resolved keychain %r to action %r', context, kc, act)
                return act
            elif kc.startswith(candidate):
                log.debug('%s:   Advancing keychain %r: %r', context, kc, '+'.join(candidate))
                action = KeyChain.ADVANCE
            elif key == 'backspace' and len(partial) > 0:
                log.debug('%s:   Reducing keychain %r: %r', context, kc, candidate[:-2])
                action = KeyChain.REDUCE
        return action

    def _keychains(self, context=ALL_CONTEXTS):
        def keychains_from(cntxt):
            for kc,action in self._actions[cntxt].items():
                if isinstance(kc, KeyChain):
                    yield (kc, action)

        if context is self.ALL_CONTEXTS:
            for cntxt in self.contexts:
                yield from keychains_from(cntxt)
        else:
            yield from keychains_from(context)

    def _started_keychains(self, context=ALL_CONTEXTS):
        keychain_partial = self._keychain_partial
        for kc,action in self._keychains(context):
            if kc.startswith(keychain_partial):
                yield (kc, action)

    def _reset_keychains(self, context=ALL_CONTEXTS):
        log.debug('%s: Resetting keychains', context)
        self._keychain_partial.clear()
        self._keychain_context = self.NO_CONTEXT
        self._run_keychain_callbacks()

    @property
    def any_keychain_started(self):
        """Whether any keychain was started"""
        return bool(self._keychain_partial)

    @property
    def current_keychain(self):
        """Tuple of currently collected keychain keys"""
        return tuple(self._keychain_partial)

    def _run_keychain_callbacks(self):
        context = self._keychain_context
        if context is self.NO_CONTEXT:
            self._keychain_callbacks.send(self, context=self.NO_CONTEXT, active_keychains=(), keys_given=())
        else:
            self._keychain_callbacks.send(self,
                                          context=context,
                                          active_keychains=self._started_keychains(context),
                                          keys_given=self.current_keychain)

    def on_keychain(self, callback):
        """
        Register callback for keychain changes

        `callback` is called when a keychain is advanced, finished or aborted
        with this KeyMap instances as a positional argument and the keyword
        arguments "active_keychains" and "keys_given".
        """
        self._keychain_callbacks.connect(callback, weak=True)

    def on_bind_unbind(self, callback):
        """
        Register callback for bind/unbind events

        `callback` is called when a key mapped or unmapped with this KeyMap
        instances as a positional argument.
        """
        self._bindunbind_callbacks.connect(callback, weak=True)

    def wrap(self, cls, callback=None, context=DEFAULT_CONTEXT):
        """
        Return widget class that passes keypress()es through `evaluate`

        callback: Custom callable for this widget class; will be called with
                  the action mapped to the key and the widget instance.
        context: A descriptive string (or any other hashable object)
        """
        if context not in self._actions:
            self._actions[context] = {}
        clsname = cls.__name__ + '_Keymapped'
        return type(clsname,
                    (Keymapped, cls),
                    {'_Keymapped__keymap': self,
                     '_Keymapped__context': context,
                     '_Keymapped__callback': staticmethod(callback)})

    _KEY_SPLIT_SPACE = re.compile(r' +')
    def mkkey(self, string):
        """
        Create Key or KeyChain instance from `string`

        Valid values for `string`:
          - A Key or KeyChain instance, which is simply returned.
          - A string that contains spaces is split at those spaces and all
            non-empty items of the resulting list are passed to KeyChain.
          - A string with '+' in it is split at those '+' and the resulting list
            is passed to KeyChain.
          - Any other string is passed to Key.
        """
        if isinstance(string, (Key, KeyChain)):
            return string
        elif ' ' in string:
            keys = (key for key in self._KEY_SPLIT_SPACE.split(string) if key)
            return KeyChain(*keys)
        elif '+' in string and len(string) >= 3:
            return KeyChain(*string.strip().split('+'))
        else:
            return Key(string)

    def keys(self, match_func=None, context=None):
        """
        Yield all keys where `match_func(key, action)` evaluates to True

        If `context` is not None, yield only keys mapped in that context.
        """
        def is_match(key, action):
            return match_func is None or match_func(key, action)

        if context is not None:
            for k,a in self._actions[context].items():
                if is_match(k, a):
                    yield k
        else:
            for context in self.contexts:
                for k,a in self._actions[context].items():
                    if is_match(k, a):
                        yield k

    def map(self, context=DEFAULT_CONTEXT):
        """Yield (key, action) tuples for keys mapped in `context`"""
        yield from self._actions[context].items()

    @property
    def contexts(self):
        """Yield existing contexts"""
        yield from self._actions



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for setting list columns"""

from . import ColumnBase


COLUMNS = {}
ALIASES = {}


class Name(ColumnBase):
    header = {'left': 'Name'}
    align = 'left'
    width = 25

    def get_value(self):
        return self.data['id']

COLUMNS['name'] = Name


class Value(ColumnBase):
    header = {'left': 'Value'}
    align = 'left'
    width = None
    wrap = 'space'

    def get_raw_value(self):
        return self.data['value']

    def get_value(self):
        value = self.data['value']
        if hasattr(value, 'prettified'):
            return value.prettified
        else:
            return value

COLUMNS['value'] = Value


class Default(ColumnBase):
    header = {'left': 'Default'}
    align = 'left'
    width = None
    wrap = 'space'

    def get_raw_value(self):
        return self.data['default']

    def get_value(self):
        value = self.data['default']
        # Paths are prettified by replacing '~' with $HOME
        if hasattr(value, 'prettified'):
            return value.prettified
        else:
            return value

COLUMNS['default'] = Default


class Description(ColumnBase):
    header = {'left': 'Description'}
    align = 'left'
    width = None
    wrap = 'space'

    def get_value(self):
        return self.data['description']

COLUMNS['description'] = Description



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid

from ..table import ColumnHeaderWidget
from . import (Style, CellWidgetBase)
from ...views.setting import COLUMNS as _COLUMNS


TUICOLUMNS = {}

class Name(_COLUMNS['name'], CellWidgetBase):
    style = Style(prefix='settinglist.name', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['name'].header),
                           style.attrs('header'))

TUICOLUMNS['name'] = Name


class Value(_COLUMNS['value'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='settinglist.value', focusable=True, extras=('header',),
                  modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['value'].header),
                           style.attrs('header'))

    def get_mode(self):
        default = self.data['default']
        value = self.data['value']
        if not self.data['id'].startswith('srv.') and value != default:
            return 'highlighted'

TUICOLUMNS['value'] = Value


class Default(_COLUMNS['default'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='settinglist.default', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['default'].header),
                           style.attrs('header'))

TUICOLUMNS['default'] = Default


class Description(_COLUMNS['description'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='settinglist.description', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['description'].header),
                           style.attrs('header'))

TUICOLUMNS['description'] = Description



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..ttypes import SmartCmpStr
from . import (BoolFilterSpec, CmpFilterSpec, Filter, FilterChain)


class SingleSettingFilter(Filter):
    DEFAULT_FILTER = 'name'

    # Filters without arguments
    BOOLEAN_FILTERS = {
        'all': BoolFilterSpec(
            lambda s: True,
            aliases=('*',),
            description='All settings'),
        'changed': BoolFilterSpec(
            lambda s: s['default'] != s['value'],
            aliases=('ch',),
            description='Settings with customized values'),
    }

    COMPARATIVE_FILTERS = {
        'name': CmpFilterSpec(
            lambda s, op, v: op(SmartCmpStr(s['id']), v),
            aliases=('n',),
            description='Match VALUE against name of setting',
            value_type=SmartCmpStr),
        'value': CmpFilterSpec(
            lambda s, op, v: op(SmartCmpStr(s['value']), v),
            aliases=('v',),
            description='Match VALUE against current value',
            value_type=SmartCmpStr),
        'default': CmpFilterSpec(
            lambda s, op, v: op(SmartCmpStr(s['default']), v),
            aliases=('def',),
            description='Match VALUE against default value',
            value_type=SmartCmpStr),
        'description': CmpFilterSpec(
            lambda s, op, v: op(SmartCmpStr(s['description']), v),
            aliases=('desc',),
            description='Match VALUE against description',
            value_type=SmartCmpStr),
    }


class SettingFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleSettingFilter
